import collections
import string
import re

__ALL__ = ['clean_sentence_fn', 'build_vocab_fn']


def clean_sentence_fn(inputs, lowercase=False, rep_num=False, rep_url=False, rm_punc=False):
    """
    The `clean_sentence_fn` is a function for cleaning texts,
    including transforming words into lowercase, replacing numbers with `<num>` token,
    replacing urls with `<url>` token, and removing punctuations.

    Parameters
    ----------
    inputs: a string or list of strings
        Each string represents a text sentence to be cleaned.
    lowercase: boolean
        If true, transform all words into lowercase. Default to be False.
    rep_num: boolean
        If true, replace numbers with `<num>` token. Default to be False.
    rep_url: boolean
        If true, replace urls with `<url>` token. Default to be False.
    rm_punc: boolean
        If true, remove all punctuation marks. Default to be False.

    Returns
    -------
    outputs: a string or list of strings
        The cleaned sentence/sentences.

    """

    # The sequential order is important.
    def clean_fn(sentence):
        if lowercase:
            sentence = sentence.strip().lower()
        # if rep_url:
        #     sentence = replace_url_fn(sentence)
        if rm_punc:
            remove_punctuation = str.maketrans("", "", string.punctuation)
            sentence = sentence.translate(remove_punctuation)
        if rep_num:
            sentence = replace_num_fn(sentence)
        return sentence

    def replace_num_fn(sentence):

        # This function checks whether a word is a digit.
        # 10: True
        # 3.21: True
        # 9:00: False
        # $10: False
        # 1e3: True
        # -1.37: True
        def is_number(s):
            try:
                float(s)
                return True
            except ValueError:
                pass
            try:
                import unicodedata
                unicodedata.numeric(s)
                return True
            except (TypeError, ValueError):
                pass
            return False

        # This function checks whether a word has digits.
        # All examples above are True.
        def has_number(s):
            return bool(re.search(r'\d', s))

        split = [word if has_number(word) is False else "<num>" for word in sentence.split()]
        return " ".join(split).strip()

    # def replace_url_fn(sentence):
    #     # to-do
    #     return sentence

    if isinstance(inputs, list):
        outputs = []
        for sent in inputs:
            outputs.append(clean_fn(sent))
    elif isinstance(inputs, str):
        outputs = clean_fn(inputs)
    else:
        raise TypeError("inputs should be a string or list of strings.")
    return outputs


def build_vocab_fn(value_list, max_vocab_size=-1, prev_list=[]):
    """
    The `build_vocab_fn` is a function for building vocabulary from data.
    It consists of all unique values of given dataset.

    Parameters
    ----------
    value_list: a list
        A list of values.
    max_vocab_size: an integer
        The maximum number of vocabulary. Default to be -1, which means no limits.
    prev_list: a list
        The prefix list of vocabulary. Usually they are manually-designated special tokens.
        Default to be an empty string.

    Returns
    -------
    word2id: a dict
        A vocabulary dict mapping from words to indexes.
    id2word: a dict
        A vocabulary dict mapping from indexes to words.

    """
    counter = collections.Counter(value_list).most_common()
    value_list, freq_list = zip(*counter)
    vocab_list = prev_list + list(value_list)

    if max_vocab_size != -1:
        vocab_list = vocab_list[:max_vocab_size]

    index_list = range(len(vocab_list))

    word2id = dict(zip(vocab_list, index_list))
    id2word = dict(zip(index_list, vocab_list))

    return word2id, id2word


def load_vocab_fn(vocab_file_path):
    """
    The `load_vocab_fn` loads existing vocabulary from `vocab_path`.

    Parameters
    ----------
    vocab_file_path: a string
        The path of existing vocabulary file ``vocab.txt``.

    Returns
    -------
    word2id: a dict
        A vocabulary dict mapping from words to indexes.
    id2word: a dict
        A vocabulary dict mapping from indexes to words.

    """
    with open(vocab_file_path, mode="r", encoding="utf-8") as file:
        vocab_list = file.read().split("\n")[:-1]

    index_list = range(len(vocab_list))
    word2id = dict(zip(vocab_list, index_list))
    id2word = dict(zip(index_list, vocab_list))

    return word2id, id2word


import tensorflow as tf
from deepdialog.utils import rnn_builder
from deepdialog.utils import attention_builder

from deepdialog.component.base import DeterministicComponent

__ALL__ = ['Decoder']

class Decoder(DeterministicComponent):
    """
    The :class:`Decoder` class defines the decoder component of generative dialogue models.
    It is implemented through a recurrent neural network.

    Parameters
    ----------
    embedding: an Tensorflow variable
        The vocabulary embedding.
    mode: string
        The working mode
    """
    def __init__(self, embedding, mode, params, decoder_keep_prob, name="decoder"):

        self._embedding = embedding
        self._mode = mode
        self._params = params
        self._decoder_keep_prob = decoder_keep_prob

        # self._input_shape = []
        # self._output_shape = []

        DeterministicComponent.__init__(self, name, [], [], tf.float32)


    def __call__(self, encoder_outputs,encoder_states, encoder_lengths, decoder_inputs, decoder_lengths):

        with tf.variable_scope("attention_decoder") as scope:

            decoder_cell, decoder_initial_state = self._prepare_decoder(
                encoder_outputs, encoder_states, encoder_lengths)

            output_layer = tf.layers.Dense(self._params.vocab_size,
                                           kernel_initializer=tf.truncated_normal_initializer(mean=0.0, stddev=0.1),
                                           _scope='decoder/dense')

            if self._mode != tf.contrib.learn.ModeKeys.INFER:  # train

                decoder_embedded_inputs = tf.nn.embedding_lookup(self._embedding, decoder_inputs)
                helper = tf.contrib.seq2seq.TrainingHelper(decoder_embedded_inputs, decoder_lengths, time_major=False)
                basic_decoder = tf.contrib.seq2seq.BasicDecoder(
                    decoder_cell, helper, decoder_initial_state, output_layer=output_layer)

                outputs, final_decoder_state, _ = tf.contrib.seq2seq.dynamic_decode(
                    basic_decoder, impute_finished=True, swap_memory=True, scope=scope)

                sample_id = outputs.sample_id
                logits = tf.identity(outputs.rnn_output)

            else:  # inference
                beam_width = self._params.beam_width
                start_tokens = tf.fill([self._params.batch_size], self._params.SOS_ID)
                end_token = self._params.EOS_ID

                decoding_length_factor = self._params.decoding_length_factor
                max_encoder_length = tf.reduce_max(encoder_lengths)
                maximum_iterations = tf.to_int32(tf.round(tf.to_float(max_encoder_length) * decoding_length_factor))

                if beam_width > 0:
                    infer_decoder = tf.contrib.seq2seq.BeamSearchDecoder(decoder_cell, self._embedding, start_tokens,
                                                                         end_token,
                                                                         initial_state=decoder_initial_state,
                                                                         beam_width=beam_width,
                                                                         output_layer=output_layer,
                                                                         length_penalty_weight=self._params.length_penalty_weight)
                else:
                    helper = tf.contrib.seq2seq.GreedyEmbeddingHelper(self._embedding, start_tokens, end_token)
                    infer_decoder = tf.contrib.seq2seq.BasicDecoder(
                        decoder_cell,
                        helper,
                        decoder_initial_state,
                        output_layer=output_layer)

                outputs, final_decoder_state, _ = tf.contrib.seq2seq.dynamic_decode(
                    infer_decoder,
                    maximum_iterations=maximum_iterations,
                    swap_memory=True,
                    scope=scope)

                if beam_width > 0:
                    logits = tf.no_op()
                    sample_id = outputs.predicted_ids
                else:
                    logits = outputs.rnn_output
                    sample_id = outputs.sample_id

            return logits, sample_id, final_decoder_state


    def _prepare_decoder(self, encoder_outputs, encoder_states, encoder_lengths):

        decoder_cell = rnn_builder.create_rnn_cell(
            self._params.cell_type, self._params.hidden_dim, self._params.num_layers,
            use_residual=self._params.use_residual, input_keep_prob=self._decoder_keep_prob)

        if self._mode == tf.contrib.learn.ModeKeys.INFER and self._params.beam_width > 0:
            batch_size = self._params.batch_size * self._params.beam_width
            decoder_initial_state = tf.contrib.seq2seq.tile_batch(encoder_states, multiplier=self._params.beam_width)
            memory = tf.contrib.seq2seq.tile_batch(encoder_outputs, multiplier=self._params.beam_width)
            memory_lengths = tf.contrib.seq2seq.tile_batch(encoder_lengths, multiplier=self._params.beam_width)
        else:
            batch_size = self._params.batch_size
            decoder_initial_state = encoder_states
            memory = encoder_outputs
            memory_lengths = encoder_lengths

        if not self._params.use_attention:
            return decoder_cell, decoder_initial_state

        attention_mechanism = attention_builder.create_attention(self._params.attention_type, self._params.hidden_dim,
                                                                 memory,
                                                                 memory_lengths)
        alignment_history = (self._mode == tf.contrib.learn.ModeKeys.INFER and self._params.beam_width == 0)
        decoder_cell = tf.contrib.seq2seq.AttentionWrapper(decoder_cell,
                                                           attention_mechanism=attention_mechanism,
                                                           attention_layer_size=self._params.hidden_dim,
                                                           alignment_history=alignment_history,
                                                           output_attention=True,
                                                           name="vanilla_attention")

        decoder_initial_state = decoder_cell.zero_state(batch_size, tf.float32).clone(
            cell_state=decoder_initial_state)

        return decoder_cell, decoder_initial_state






import os
import time
import tensorflow as tf
import numpy as np
from deepdialog.model.vhred import vhred_wrapper
from deepdialog.data_utils.dataset import HierarchicalDataset
from deepdialog.data_utils.data_loader import HierarchicalDataLoader
from deepdialog.model import model_helper


class DataLoaderConfig:
    batch_size = 8
    train_valid_split = 0.6
    valid_test_split = 0.8
    tokenized = True
    max_dialog_size = 3
    max_utterance_length = 8


class DatasetConfig:
    utterance_separator = "__eou__"
    dataset_name = "dailydialog"
    max_vocab_size = 20000
    max_dialog_size = 3


class ModelConfig:
    hidden_units = 3
    ctx_enc_dim = 2
    embed_size = 4
    latent_size = 2
    init_lr = 0.001
    lr_decay = 0.5
    encoder_type = "uni"
    cell_type = "gru"
    num_layers = 1
    use_residual = False
    batch_size = 8
    use_attention = False
    attention_type = "luong"
    vocab_size = 20000
    encoder_dropout_rate = 0.0
    decoder_dropout_rate = 0.0
    op = "adam"
    grad_clip = 5.0
    grad_noise = 0.0
    epochs = 3
    full_kl_step = 10000


class VHREDRunner(object):
    def __init__(self):
        dataset_config = DatasetConfig()
        self.dataset = dataset = HierarchicalDataset(dataset_config)

        cleaned_texts = dataset.get_cleaned_texts()
        label_list = dataset.get_raw_label_list()
        word2id = dataset.get_word2id()
        id2word = dataset.get_id2word()
        label2id_list = dataset.get_label2id_list()
        id2label_list = dataset.get_id2label_list()
        label_name_list = dataset.get_label_name_list()

        data_loader_config = DataLoaderConfig()
        self.dataloader = dataloader = HierarchicalDataLoader(
            data_loader_config, cleaned_texts, label_list, word2id, id2word,
            label2id_list, id2label_list, label_name_list)

        self.model_config = ModelConfig()
        self.test_config = ModelConfig()

        self.out_dir = "../../vhred_save"
        self.model_dir = "../../vhred_save"

        self.eval_model = vhred_wrapper.create_vhred_test_model(
            dataset, dataloader, self.test_config, scope="vhred")
        self.train_model = vhred_wrapper.create_vhred_train_model(
            dataset, dataloader, self.model_config, scope="vhred")

        self.eval_sess = tf.Session(graph=self.eval_model.graph)
        self.train_sess = tf.Session(graph=self.train_model.graph)

    def train(self):
        with self.train_model.graph.as_default():
            loaded_train_model, global_step = model_helper.create_or_load_model(
                self.train_model.model, self.model_dir, self.train_sess, "train")
        """
        summary_writer = tf.summary.FileWriter(
            os.path.join(self.out_dir, "train_log"), self.train_model.graph)
        """
        for epoch in range(self.model_config.epochs):
            start_time = time.time()
            # print(epoch)
            train_loss_list = []
            elbo_list = []
            rc_loss_list = []
            kl_loss_list = []
            for _, batch_data in enumerate(self.train_model.iterator()):
                step_result = loaded_train_model.train(self.train_sess, batch_data)

                _, train_loss, elbo, rc_loss, kl_loss, predict_count, \
                train_summary, self.global_t, self.learning_rate = step_result
                train_loss_list.append(train_loss)
                elbo_list.append(elbo)
                rc_loss_list.append(rc_loss)
                kl_loss_list.append(kl_loss)
            print("Epoch %d/%d" % (epoch + 1, self.model_config.epochs))
            print("avg train loss: %f, elbo: %f, avg rc loss: %f, avg kl loss: %f"
                  % (np.mean(train_loss_list), np.mean(elbo_list), np.mean(rc_loss_list), np.mean(kl_loss_list)))
            print()
        print("train done")
        loaded_train_model.saver.save(
            self.train_sess,
            os.path.join(self.out_dir, "vhred.ckpt"),
            global_step=global_step)

    def test(self):
        with self.eval_model.graph.as_default():
            loaded_eval_model, global_step = model_helper.create_or_load_model(
                self.eval_model.model, self.model_dir, self.eval_sess, "eval")
        """
        summary_writer = tf.summary.FileWriter(
            os.path.join(self.out_dir, "train_log"), self.train_model.graph)
        """
        eval_loss_list = []
        elbo_list = []
        rc_loss_list = []
        kl_loss_list = []
        for _, batch_data in enumerate(self.eval_model.iterator()):
            eval_loss, elbo, rc_loss, kl_loss, predict_count = loaded_eval_model.test(self.eval_sess,
                                                                                      batch_data)
            eval_loss_list.append(eval_loss)
            elbo_list.append(elbo)
            rc_loss_list.append(rc_loss)
            kl_loss_list.append(kl_loss)
        print("avg eval loss: %f, elbo: %f, avg rc loss: %f, avg kl loss: %f"
              % (np.mean(eval_loss_list), np.mean(elbo_list), np.mean(rc_loss_list), np.mean(kl_loss_list)))
        print("eval done")


    def interactive(self):
        pass



from abc import abstractmethod

import tensorflow as tf

__ALL__=['BaseComponent', 'DeterministicComponent', 'StochasticComponent']


class BaseComponent(object):
    """
    The :class:`Component` is the base class for constitutional components of generative dialogue models.
    It supports batch inputs and batch outputs.
    Each component consists of a series of operations and targets a specific function.
    The components could be categorized into either deterministic or stochastic types.

    The typical input shape for :class:`Component` is like ``[batch_shape,input_shape]``,
    where ``input_shape`` represents the shape of non-batch input parameter,
    ``batch_shape`` represents how many independent inputs are fed into the :class:`Component`.

    The outputs are usually of shape ``[batch_shape, value_shape]``,
    where ``value_shape`` is the non-batch value shape of the :class:`Component`.
    Particularly, for stochastic components,
    the outputs could be of shape ``[n_sample, batch_shape, value_shape]``.
    The first additional axis is omitted only when passed n_samples is None (by default),
    in which case one sample is generated.

    The parameter dtype represents type of outputs.
    It is automatically determined from parameter types.
    And dtype must be among int16, int32, int64, float16, float32 and float64.
    When two or more parameters are tensors and they have different type, `TypeError` will be raised.

    Parameters
    ----------
    name: string
        The name of the component.
    input_shape:
        The tensor shape of inputs.
    output_shape:
        The tensor shape of outputs.
    category:
        The category of the component. either "deterministic" or "stochastic".
    dtype:
        The value type of outputs from the component.
    """
    def __init__(self, name, input_shape, output_shape, category, dtype):
        self._name = name
        self._input_shape = input_shape
        self._output_shape = output_shape
        self._dtype = dtype
        self._category = category

        if category not in ("deterministic", "stochastic"):
            raise ValueError("category can be either 'deterministic' or 'stochastic'.")

    @property
    def name(self):
        """The name of the component."""
        return self._name

    @property
    def input_shape(self):
        """The shape of the inputs into the component. A TensorShape object."""
        return self._name

    @property
    def output_shape(self):
        """The shape of the outputs from the component. A TensorShape object."""
        return self._output_shape

    @property
    def dtype(self):
        """The value type of outputs from the component."""
        return self._dtype

    @property
    def category(self):
        """The category of the component. It could be either `Stochastic` or `Deterministic`."""
        return self._category


class DeterministicComponent(BaseComponent):
    """
    The :class:`DeterministicComponent` is class for deterministic components of generative dialogue
    models.
    """
    def __init__(self, name, input_shape, output_shape, dtype):
        BaseComponent.__init__(self, name, input_shape, output_shape, "deterministic", dtype)

    @abstractmethod
    def __call__(self, *args, **kwargs):
        pass

class StochasticComponent(BaseComponent):
    """
    The :class:`StochasticComponent` is class for stochastic components of generative dialogue
    models.
    It defines a variable distribution and supports stochastic sampling operation.

    Parameters
    ----------
    name: string
        The name of the component.
    n_samples: integer
        The number of samples to be generated.
    """
    def __init__(self, name, n_samples=None, dtype=tf.float32):
        BaseComponent.__init__(self, name, [], [], "stochastic", dtype)

        self._n_samples = n_samples
        self._distribution = None

    @abstractmethod
    def __call__(self, *args, **kwargs):
        pass

    @property
    def n_samples(self):
        """The number of samples to be generated from the component."""
        return self._n_samples

    def sample(self, n_sample=None):
        """Sampling any number of instances from the component."""
        return self._distribution.sample(n_sample)





import tensorflow as tf
from deepdialog.utils import rnn_builder

from deepdialog.component.base import DeterministicComponent

__ALL__ = ['Encoder', "ContextEncoder"]


class Encoder(DeterministicComponent):
    """
    The :class:`Encoder` class is the encoder component of generative dialogue models,
    which encodes input sentences into real-valued vectors.
    The real-valued vectors are usually considered the summary of sentences.
    The :class:`Encoder` class is implemented through a recurrent neural network.

    The typical shape of inputs is like ``[batch_size, seq_len, embed_size]``.
    The shape of output vectors is like ``[batch_size, hidden_dim]``.
    """
    def __init__(self, embedding, params, input_keep_prob, name="vanilla_encoder"):

        self._input_shape = tf.TensorShape([None, params.seq_len])
        self._output_shape = tf.TensorShape([None, params.hidden_dim])
        self._embedding = embedding
        self._params = params
        self._input_keep_prob = input_keep_prob

        DeterministicComponent.__init__(self, name, self._input_shape, self._output_shape, tf.float32)


    def __call__(self, encoder_inputs, encoder_lengths):

        if len(encoder_inputs.get_shape()) >= 3 and \
                (len(encoder_lengths.get_shape()) + 1 == len(encoder_inputs.get_shape())):
            encoder_inputs = tf.reshape(encoder_inputs, [-1, tf.shape(encoder_inputs)[-1]])
            encoder_lengths = tf.reshape(encoder_lengths, [-1])

        with tf.variable_scope("encoder") as scope:
            encoder_embedded_inputs = tf.nn.embedding_lookup(self._embedding, encoder_inputs)
            # [batch_size, seq_len, embed_size]

        if self._params.encoder_type == "uni":
            encoder_cell = rnn_builder.create_rnn_cell(
                self._params.cell_type, self._params.hidden_dim, self._params.num_layers,
                use_residual=self._params.use_residual, input_keep_prob=self._input_keep_prob)

            encoder_outputs, encoder_states = tf.nn.dynamic_rnn(
                cell=encoder_cell, inputs=encoder_embedded_inputs,
                sequence_length=encoder_lengths, dtype=tf.float32,
                swap_memory=True)

            # gru: [batch_size, seq_len, hidden_size] [batch_size, hidden_size]
            # lstm: [batch_size, hidden_size] * num_layers
            return encoder_outputs, encoder_states

        elif self._params.encoder_type == "bi":

            fw_cell = rnn_builder.create_rnn_cell(
                self._params.cell_type, self._params.hidden_units, self._params.num_layers,
                use_residual=self._params.use_residual, input_keep_prob=self._input_keep_prob)

            bw_cell = rnn_builder.create_rnn_cell(
                self._params.cell_type, self._params.hidden_units, self._params.num_layers,
                use_residual=self._params.use_residual, input_keep_prob=self._input_keep_prob)

            encoder_outputs, encoder_states = tf.nn.bidirectional_dynamic_rnn(
                cell_fw=fw_cell, cell_bw=bw_cell, inputs=encoder_embedded_inputs,
                dtype=tf.float32, sequence_length=encoder_lengths, swap_memory=True)
            return encoder_outputs, encoder_states

        else:
            raise ValueError("Unknown encoder_type Error: %s" % self._params.encoder_type)



class ContextEncoder(DeterministicComponent):
    """
    The :class:`ContextEncoder` class is the context encoder component of generative dialogue models,
    which encodes dialogue contexts into real-valued vectors.
    The real-valued vectors are usually considered the summary of multi-turn utterances.
    The :class:`ContextEncoder` class is implemented through a recurrent neural network.

    The typical shape of inputs is like ``[batch_size, seq_len]``.
    The shape of output vectors is like ``[batch_size, hidden_dim]``.
    """
    def __init__(self, params, input_keep_prob, name="context_encoder"):
        self._input_shape = tf.TensorShape([None, params.seq_len])
        self._output_shape = tf.TensorShape([None, params.hidden_dim])
        self._params = params
        self._input_keep_prob = input_keep_prob

        DeterministicComponent.__init__(self, name, self._input_shape, self._output_shape, tf.float32)


    def __call__(self, inputs, encoder_lengths):

        with tf.variable_scope("context_encoder") as scope:
            if self._params.encoder_type == "uni":
                encoder_cell = rnn_builder.create_rnn_cell(
                    self._params.cell_type, self._params.ctx_enc_dim, self._params.num_layers,
                    use_residual=self._params.use_residual, input_keep_prob=self._input_keep_prob)

                encoder_outputs, encoder_states = tf.nn.dynamic_rnn(
                    cell=encoder_cell, inputs=inputs,
                    sequence_length=encoder_lengths, dtype=tf.float32,
                    swap_memory=True)

                # gru: [batch_size, seq_len, hidden_size] [batch_size, hidden_size]
                # lstm: [batch_size, hidden_size] * num_layers
                return encoder_outputs, encoder_states

            elif self._params.encoder_type == "bi":

                fw_cell = rnn_builder.create_rnn_cell(
                    self._params.cell_type, self._params.hidden_units, self._params.num_layers,
                    use_residual=self._params.use_residual, input_keep_prob=self._input_keep_prob)

                bw_cell = rnn_builder.create_rnn_cell(
                    self._params.cell_type, self._params.hidden_units, self._params.num_layers,
                    use_residual=self._params.use_residual, input_keep_prob=self._input_keep_prob)

                encoder_outputs, encoder_states = tf.nn.bidirectional_dynamic_rnn(
                    cell_fw=fw_cell, cell_bw=bw_cell, inputs=inputs,
                    dtype=tf.float32, sequence_length=encoder_lengths, swap_memory=True)

                return encoder_outputs, encoder_states

            else:
                raise ValueError("Unknown encoder_type Error: %s" % self._params.encoder_type)


from abc import abstractmethod

import tensorflow as tf
import numpy as np


__ALL__ = ["MultiVarDiagonalGaussian", "BaseDistribution"]

class BaseDistribution(object):
    """
    The :class:`BaseDistribution` class is the base class for various probabilistic distributions.
    It support sampling from the distribution, evaluating (log-)probabilities at given values,
    and computing the kl-divergence between this distribution and a given distribution.
    """
    def __init__(self, name):
        self._name = name

    @property
    def name(self):
        return self._name

    @abstractmethod
    def sample(self):
        pass

    @abstractmethod
    def log_prob(self, inputs):
        pass

    @abstractmethod
    def prob(self, inputs):
        pass

    @abstractmethod
    def kl_divergence(self, dist):
        pass


class MultiVarDiagonalGaussian(BaseDistribution):
    """
    The :class:`DiagonalGaussian` class is the multivariate Gaussian distribution with diagonal covariance.

    It supports batch inputs, generating batches of samples, evaluating (log-)probabilities at batches of
    given values, and computing the KL-divergence between itself and another given distribution.

    The typical shape for distribution parameters is ``batch_shape + value_shape``.
    The ``batch_shape`` represents how many independent inputs are fed into the distribution at the same time.
    The ``value_shape`` represents the shape of non-batch values of this distribution.
    Both ``mean`` and ``log_variance`` share the same shape.

    Samples generated are of shape ``([n_samples]+) batch_shape + value_shape``.
    The first additional axis omitted only when passed n_samples is None (by default), in which case one sample is
    generated.
    For a uni-variate distribution, its value_shape is ``TensorShape([])``.
    For a multi-variate distribution, its value_shape is ``TensorShape([N])``, where N is an integer.

    When evaluating (log-)probabilities at given values, the given values should be of the shape
    ``([n_samples]+) batch_shape + value_shape``.
    The returned Tensor has shape ``([n_samples]+) batch_shape``.

    When computing KL-divergence given another distribution,
    the given distribution should have the same ``batch_shape`` and ``value_shape``.
    The returned Tensor has shape ``batch_shape``.

    Both parameters and samples have the same type, which equals ``dtype``.
    When two or more have different type, TypeError will be raised.

    Parameters
    ----------
    mean:
        the mean value of the distribution.
    log_variance:
        The diagonal covariance value of the distribution.
    dtype:
        The type of samples/parameters of the distribution.
    """
    def __init__(self, mean, log_variance, dtype, name):
        BaseDistribution.__init__(self, name)

        self._log_variance = tf.convert_to_tensor(log_variance)
        self._mean = tf.convert_to_tensor(mean)
        self._dtype = dtype

        # check shape
        self._mean.get_shape().assert_is_compatible_with(self._log_variance.get_shape())
        self._batch_shape = self._mean.get_shape()[:-1]
        self._value_shape = self._mean.get_shape()[-1]


    @property
    def mean(self):
        """The mean value of the distribution. It has shape ``batch_shape + value_shape``."""
        return self._mean

    @property
    def log_variance(self):
        """The diagonal covariance value of the distribution. It has shape ``batch_shape + value_shape``."""
        return self._log_variance

    @property
    def dtype(self):
        """The type of samples/parameters of the distribution."""
        return self._dtype

    @property
    def batch_shape(self):
        """
        The shape showing how many independent inputs are fed into the distribution at the same time.

        Returns
        -------
        batch_shape: A TensorShape object.
        """
        return self._batch_shape

    @property
    def value_shape(self):
        """
        The shape showing the non-batch values of this distribution.

        Returns
        -------
        value_shape: A TensorShape object.
        """
        return self._value_shape

    def _sample(self):
        """
        Generate one sample from the distribution.

        Returns
        -------
            A sample of shape ``batch_shape + value_shape``.
        """
        epsilon = tf.random_normal(tf.shape(self._log_variance), name="epsilon")
        std = tf.exp(0.5 * self._log_variance)
        z = self._mean + tf.multiply(std, epsilon)
        return z

    def sample(self, n_samples=None):
        """
        Generate multiple samples from the distribution.

        Parameters
        ----------
        n_samples: an integer
            The number of samples.

        Returns
        -------
        Samples: a Tensor
            Multiple samples of shape ``[n_samples] + batch_shape + value_shape``.
        """
        if n_samples is None:
            return self._sample()
        ret = []
        for _ in range(n_samples):
            ret.append(tf.expand_dims(self._sample(), axis=0))
        return tf.concat(ret, 0)


    def log_prob(self, inputs):
        """
        Evaluating the log-probabilities of the distribution at given values.

        Parameters
        ----------
        inputs: a Tensor
            Given values of shape ``([n_samples] +) batch_shape + value_shape``.
        """
        return -0.5 * tf.reduce_sum(
            tf.log(2 * np.pi) + self._log_variance +
            tf.div(tf.pow((inputs - self._mean), 2), tf.exp(self._log_variance)),
                                    reduction_indices=-1)

    def prob(self, inputs):
        """
        Evaluating the probabilities of the distribution at given values.

        Parameters
        ----------
        inputs: a Tensor
            Given values of shape ``([n_samples] +) batch_shape + value_shape``.
        """
        raise NotImplementedError()


    def kl_divergence(self, dist):
        """
        Computing KL[current_distribution || dist].

        Parameters
        ----------
        dist: a BaseDistribution object
            Another given distribution.
        Returns
        -------
        KLD: a Tensor
            The KL divergence between current distribution and the given distribution.
        """
        given_dist = self._check_dist_shape(dist)

        prior_logvar = given_dist.log_variance
        prior_mu = given_dist.mean

        kld = -0.5 * tf.reduce_sum(1 + (self._log_variance - prior_logvar)
                               - tf.div(tf.pow(prior_mu - self._mean, 2), tf.exp(prior_logvar))
                               - tf.div(tf.exp(self._log_variance), tf.exp(prior_logvar)), reduction_indices=-1)
        return kld


    def _check_dist_shape(self, dist):
        """
        Check whether the given distribution shares the same shape as current distribution.

        Parameters
        ----------
        dist: a BaseDistribution object
            Another given distribution.
        """
        logvar = dist.log_variance
        mean = dist.mean
        err_msg = "The given distribution should match 'batch_shape + value_shape' of current distribution."
        try:
            logvar.get_shape().assert_is_compatible_with(mean.get_shape())
            logvar.get_shape().assert_is_compatible_with(self.log_variance.get_shape())
        except ValueError:
            raise ValueError(
                err_msg + "(given: {} and {} | current: {})".format(
                        logvar.get_shape(), mean.get_shape(), self.log_variance.get_shape())
            )
        return dist


from nltk.translate.bleu_score import sentence_bleu
import numpy as np
from nltk.util import ngrams
import math
# import tensorflow as tf


def BLEU(source, target):
    """

    Calculate the BLEU/BLEU-1/BLEU-2/BLEU-3/BLEU-4 score.

    BLEU (bilingual evaluation understudy) is an algorithm for evaluating the quality of
    text which has been machine-translated from one natural language to another, and is
    also adopted in dialog system.
    Quality is considered to be the correspondence between a machine's response and that of
    a human: "the closer a machine response is to a professional human response,
    the better it is" – this is the central idea behind BLEU.

    The essence of BLEU is to calculate the frequency of the co-occurrence n-gram of two
    sentences, but some techniques are used in the calculation process.

    The following is a simple example for uni-gram:

    candidate: the the the the the the the

    reference1: the cat is on the mat

    reference2: there is a cat on the mat

    The BLEU can be calculated as:

    :math:`BLEU = RefCount / Count`

    :math:`RefCount` is the number of uni-grams in candidate that appear in references.

    :math:`Count` is the number of uni-grams in candidate.

    In this case, BLEU = 7/7 = 1.

    But this result is unreasonable since there is only one uni-gram "the" in candidate
    that is stupid but with a high score. What causes the bad result is that we
    calculate the molecule incorrectly. The improvement is as follows:

    :math:`CountClip_i^j = min(Count_i, RefCount_i^j)`

    :math:`CountClip_i = max(CountClip_i^j), j=1,2,3......`

    :math:`BLEU = sum(CountClip_i)/Count, i=1,2,3......`

    :math:`Count_i` is the number of i-th uni-gram in candidate.

    :math:`CountClip_i^j` is the truncated number of i-th uni-gram in candidate that appears in j-th reference.

    :math:`CountClip_i` is the final truncated number of i-th uni-gram that appears in all references.

    In this case, BLEU = 2/7.

    In the above we have been talking about calculating for uni-gram, now we extend to n-gram in another example.

    candidate: the cat sat on the mat

    reference: the cat is on the mat

    :math:`BLEU-n = sum(CountClip_i(n-gram))/Count(n-gram)`

    In this case, we can get BLEU-1/2/3/4:

    BLEU-1 = 5/6

    BLEU-2 = 3/5

    BLEU-3 = 1/4

    BLEU-4 = 0/3

    However, there is still a problem that is a shorter candidate tend to have a
    higher BLEU score. So we have to introduce a penalty.

    :math:`BP = 1, if c > r`

    :math:`BP = exp(1-r/c), if c <= r`

    where c is the length of candidate, and r is the length of reference.

    The final expression of BLEU with penalty is:

    :math:`BLEU = BP * exp(sum(W_n * log(BLEU-n)))`

    where :math:`W_n` is the weights for n-gram.

    References :
        https://machinelearningmastery.com/calculate-bleu-score-for-text-python/
        https://www.aclweb.org/anthology/P02-1040.pdf

    Parameters
    ----------
    source : list
        the element is a String that is an utterance
    target : list
        the reference to source, the element is a String that is an utterance

    Returns
    -------
    (avg_bleu, avg_bleu_1, avg_bleu_2, avg_bleu_3, avg_bleu_4) : tuple
        a tuple consisting of 4 element: (BLEU-1, BLEU-2, BLEU-3, BLEU-4), the element is float
    """

    assert len(source) == len(target)

    avg_bleu_1 = 0
    avg_bleu_2 = 0
    avg_bleu_3 = 0
    avg_bleu_4 = 0
    avg_bleu = 0
    num_refs = len(target)
    for i in range(num_refs):
        reference = [target[i].lower().split()]
        candidate = source[i].lower().split()
        # todo weights
        bleu = sentence_bleu(reference, candidate, weights=(0.25, 0.25, 0.25, 0.25))
        bleu_1 = sentence_bleu(reference, candidate, weights=(1, 0, 0, 0))
        bleu_2 = sentence_bleu(reference, candidate, weights=(0, 1, 0, 0))
        bleu_3 = sentence_bleu(reference, candidate, weights=(0, 0, 1, 0))
        bleu_4 = sentence_bleu(reference, candidate, weights=(0, 0, 0, 1))
        avg_bleu += bleu
        avg_bleu_1 += bleu_1
        avg_bleu_2 += bleu_2
        avg_bleu_3 += bleu_3
        avg_bleu_4 += bleu_4

    avg_bleu = avg_bleu / len(source)
    avg_bleu_1 = avg_bleu_1 / len(source)
    avg_bleu_2 = avg_bleu_2 / len(source)
    avg_bleu_3 = avg_bleu_3 / len(source)
    avg_bleu_4 = avg_bleu_4 / len(source)

    return avg_bleu, avg_bleu_1, avg_bleu_2, avg_bleu_3, avg_bleu_4


def embedding_average(source, target, word2vec):
    """
    Calculate the embedding_average score.

    Embedding_average reflects the similarity between two sentences by cosine similarity.

    Firstly, all words in two sentences are transformed from string to vectors. Then an average
    value is calculated in every dimension among words. Now two sentences have been
    transformed into two vectors that are finally used to calculated the cosine similarity.

    References :
        https://github.com/chenhongshen/HVMN/blob/master/model/Evaluation/embedding_metrics.py

    Parameters
    ----------
    source : list
        the element is a String that is an utterance
    target : list
        the reference to source, the element is a String that is an utterance
    word2vec : dict
        a dict mapping a word to a vector

    Returns
    -------
    (embedding_average_score, standard_deviation, 1.96*standard_deviation) : tuple
        a tuple consisting of 3 element:(embedding_average_score, standard_deviation, 1.96*standard_deviation)
    """
    assert len(source) == len(target)
    dim = 300
    for _, vec in word2vec.items():
        dim = len(vec)
        break
    scores = []
    for i in range(min(len(source), len(target))):
        tokens1 = source[i].strip().split()
        tokens2 = target[i].strip().split()
        X = np.zeros((dim,))
        for tok in tokens1:
            if tok in word2vec:
                X += word2vec[tok]
        Y = np.zeros((dim,))
        for tok in tokens2:
            if tok in word2vec:
                Y += word2vec[tok]

        # if none of the words have embeddings in source, count result as zero
        if np.linalg.norm(X) < 0.00000000001:
            scores.append(0)
            continue

        # if none of the words in target have embeddings, skip
        if np.linalg.norm(Y) < 0.00000000001:
            continue

        X = np.array(X) / np.linalg.norm(X)
        Y = np.array(Y) / np.linalg.norm(Y)
        o = np.dot(X, Y) / np.linalg.norm(X) / np.linalg.norm(Y)

        scores.append(o)

    scores = np.asarray(scores)
    return np.mean(scores), np.std(scores), 1.96 * np.std(scores)


def embedding_extreme(source, target, word2vec):
    """
    Calculate the embedding_extreme score.

    Embedding_extreme is another algorithm that reflects the similarity between two sentences by cosine similarity.

    Firstly, all words in two sentences are transformed from string to vectors. Then the
    maximum absolute value is selected in every dimension among words. Now two sentences have been
    transformed into two vectors that are finally used to calculated the cosine similarity.

    References :
        https://github.com/chenhongshen/HVMN/blob/master/model/Evaluation/embedding_metrics.py

    Parameters
    ----------
    source : list
        the element is a String that is an utterance
    target : list
        the reference to source, the element is a String that is an utterance
    word2vec : dict
        a dict mapping a word to a vector

    Returns
    -------
    (embedding_extreme_score, standard_deviation, 1.96*standard_deviation) : tuple
        a tuple consisting of 3 element:(embedding_extreme_score, standard_deviation, 1.96*standard_deviation)
    """
    # todo word2vec load
    assert len(source) == len(target)
    scores = []
    for i in range(min(len(source), len(target))):
        tokens1 = source[i].strip().split(" ")
        tokens2 = target[i].strip().split(" ")
        X = []
        for tok in tokens1:
            if tok in word2vec:
                X.append(word2vec[tok])
        Y = []
        for tok in tokens2:
            if tok in word2vec:
                Y.append(word2vec[tok])

        # if none of the words have embeddings in source, count result as zero
        if np.linalg.norm(X) < 0.00000000001:
            scores.append(0)
            continue

        # if none of the words in target have embeddings, skip
        if np.linalg.norm(Y) < 0.00000000001:
            continue

        xmax = np.max(X, 0)  # get positive max
        xmin = np.min(X, 0)  # get abs of min
        xtrema = []
        for i in range(len(xmax)):
            if np.abs(xmin[i]) > xmax[i]:
                xtrema.append(xmin[i])
            else:
                xtrema.append(xmax[i])
        X = np.array(xtrema)  # get extrema

        ymax = np.max(Y, 0)
        ymin = np.min(Y, 0)
        ytrema = []
        for i in range(len(ymax)):
            if np.abs(ymin[i]) > ymax[i]:
                ytrema.append(ymin[i])
            else:
                ytrema.append(ymax[i])
        Y = np.array(ytrema)

        o = np.dot(X, Y) / np.linalg.norm(X) / np.linalg.norm(Y)

        scores.append(o)

    scores = np.asarray(scores)
    return np.mean(scores), np.std(scores), 1.96 * np.std(scores)


def embedding_greedy(source, target, word2vec):
    """
    Calculate the embedding_greedy score.

    Embedding_greedy is another algorithm that reflects the similarity between two sentences by cosine similarity.

    Firstly, all words in two sentences are transformed from string to vectors. Secondly, for
    the i-th word in the first sentence, we calculate the maximum cosine similarity between
    it and all words in the second sentence. Now the first sentence has m cosine similarity
    (m is the length of the first sentence). Thirdly, an average value is calculated in these
    cosine similarity. Then we do the same operation for the second sentence. Now there are
    two cosine similarity and the mean of them is the final result.

    References :
        https://github.com/chenhongshen/HVMN/blob/master/model/Evaluation/embedding_metrics.py

    Parameters
    ----------
    source : list
        the element is a String that is an utterance
    target : list
        the reference to source, the element is a String that is an utterance
    word2vec : dict
        a dict mapping a word to a vector

    Returns
    -------
    (embedding_greedy_score, standard_deviation, 1.96*standard_deviation) : tuple
        a tuple consisting of 3 element:(embedding_greedy_score, standard_deviation, 1.96*standard_deviation)
    """
    assert len(source) == len(target)
    embedding_greedy_score = []
    for i in range(len(source)):
        sentence1 = source[i].strip().split()
        sentence2 = target[i].strip().split()

        greedy_1 = greedy_score(sentence1, sentence2, word2vec)
        greedy_2 = greedy_score(sentence2, sentence1, word2vec)

        embedding_greedy_score.append((greedy_1 + greedy_2) / 2)

    return np.mean(embedding_greedy_score), \
           np.std(embedding_greedy_score), \
           1.96 * np.std(embedding_greedy_score)


def greedy_score(source, target, word2vec):
    """
    The function to be called by embedding_greedy() function and actually executes
    the algorithm given two sentence. The embedding_greedy() function calculate the
    mean and variance score among a batch data.

    Parameters
    ----------
    source : list
        the element is a String that is a word
    target : list
        the reference to source, the element is a String that is a word
    word2vec : dict
        a dict mapping a word to a vector

    Returns
    -------
    greedy_score : float
        the greedy score
    """
    cosine_list = []
    word_count = 0

    Y = []
    for tok in target:
        if tok in word2vec:
            Y.append(word2vec[tok])
    if len(Y) == 0:
        return 0.0

    for tok in source:
        if tok in word2vec:
            i_vector = word2vec[tok]
        else:
            continue

        word_count += 1
        cosine_list.append(max_cosine(i_vector, Y))

    if word_count == 0:
        return 0.0

    score = sum(cosine_list) / word_count
    return score


def max_cosine(vector, vector_list):
    """
    The function to help to calculate the embedding_greedy. It executes the second
    operation described in embedding_greedy() function.

    Parameters
    ----------
    vector : list
        the element is a float
    vector_list : list
        With 2 dimension, the element in axis0 is a list, the element in axis1 is a float

    Returns
    -------
    max_cosine : float
        the max cosine
    """
    max_value = 1e-10
    for i in vector_list:
        cosine_similarity = np.dot(vector, i) / np.linalg.norm(vector) / np.linalg.norm(i)
        max_value = max(cosine_similarity, max_value)
    return max_value

def WER(source, target):
    """
    Calculate the word error rate.

    Word error rate (WER) is a common metric of the performance of a speech recognition
    or machine translation system, and is also adopted in dialog system. It actually is
    calculated as the edit distance divided by the length of reference.

    The edit distance is  edit distance is a way of quantifying how dissimilar two
    strings (e.g., words) are to one another by counting the minimum number of operations
    required to transform one string into the other. For WER, operations are removal,
    insertion, or substitution of a word in the sentence.


    References :
        https://blog.csdn.net/quheDiegooo/article/details/56834417
        https://www.csie.ntu.edu.tw/~b93076/Computation%20of%20Normalized%20Edit%20Distance%20and%20Applications.pdf

    Parameters
    ----------
    source : list
        the element is a String that is an utterance
    target : list
        the reference to source, the element is a String that is an utterance

    Returns
    -------
    (word_error_rate, standard_deviation, 1.96*standard_deviation) : tuple
    """
    assert len(source) == len(target)
    wer_list = []
    for i in range(len(source)):
        s = source[i].strip().split()
        t = target[i].strip().split()
        s_len = len(s)
        t_len = len(t)
        edit_dis = edit_distance(s, t)
        wer_list.append(float(edit_dis) / t_len)
        """
        not_equal = 0
        for j in range(t_len):
            if j < s_len:
                not_equal += 0 if s[j] == t[j] else 1
            else:
                not_equal += 1
        wer_list.append(float(not_equal) / t_len)
        """
    wers = np.asarray(wer_list)

    return np.mean(wers), np.std(wers), 1.96 * np.std(wers)


def edit_distance(source, target):
    """
    Calculate the edit distance between two lists or two Strings.

    The edit distance is  edit distance is a way of quantifying how dissimilar two
    strings (e.g., words) are to one another by counting the minimum number of operations
    required to transform one string into the other. For WER, operations are removal,
    insertion, or substitution of a word in the sentence.

    References :
        https://leetcode-cn.com/problems/edit-distance/solution/bian-ji-ju-chi-by-leetcode/

    Parameters
    ----------
    source : list or String
    target : list or String

    Returns
    -------
    edit_distance : int
    """
    n = len(source)
    m = len(target)

    # if one of the utterance is empty
    if n * m == 0:
        return n + m

    # array to store the convertion history
    d = [[0] * (m + 1) for _ in range(n + 1)]

    # init boundaries
    for i in range(n + 1):
        d[i][0] = i
    for j in range(m + 1):
        d[0][j] = j

    # DP compute
    for i in range(1, n + 1):
        for j in range(1, m + 1):
            left = d[i - 1][j] + 1
            down = d[i][j - 1] + 1
            left_down = d[i - 1][j - 1]
            if source[i - 1] != target[j - 1]:
                left_down += 1
            d[i][j] = min(left, down, left_down)

    return d[n][m]


import time
from abc import abstractmethod
import numpy as np
import re
import tensorflow as tf



class AbstractModel(object):
    def __init__(self, params, mode, scope=None):

        self.params = params
        self.mode = mode
        self.scope = scope

        self.embeddings = None
        self.params_size = None
        self.train_ops = None
        self.learning_rate = tf.Variable(float(params.init_lr), trainable=False, name="learning_rate")
        self.global_step = tf.Variable(0, dtype=tf.int32, name="global_step")
        self.encoder_keep_prob = 1.0
        self.decoder_keep_prob = 1.0


        self.loss = None
        self.other_losses = {}

        self.init_embeddings(params.vocab_size, params.embed_dim)
        self.get_keep_probs(mode, params.encoder_dropout_rate, params.decoder_dropout_rate)
        self.prepare_io()
        self.build_graph()
        self.prepare_loss()

        if self.scope is None:
            self.tvars = tf.trainable_variables()
        else:
            self.tvars = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, scope=self.scope)

        if self.mode == tf.contrib.learn.ModeKeys.TRAIN:
            self.prepare_optimizer(self.tvars, self.loss,
                                   params.optimizer_type,
                                   params.init_lr,
                                   params.grad_clip,
                                   params.grad_noise)

        self.saver = tf.train.Saver(tf.global_variables())

    def init_embeddings(self, vocab_size, embed_dim, scope=None):
        with tf.variable_scope(scope or "embeddings", dtype=tf.float32):
            embeddings = tf.get_variable("trainable_embeddings",
                                         shape=[vocab_size, embed_dim],
                                         trainable=True,
                                         dtype=tf.float32)
            embedding_mask = tf.constant(
                [0 if i == 0 else 1 for i in range(vocab_size)], dtype=tf.float32, shape=[vocab_size, 1])
            self.embeddings = embeddings * embedding_mask

    def get_keep_probs(self, mode, encoder_dropout_rate, decoder_dropout_rate):
        if mode == tf.contrib.learn.ModeKeys.TRAIN:
            encoder_keep_prob = 1.0 - encoder_dropout_rate
            decoder_keep_prob = 1.0 - decoder_dropout_rate
        else:
            encoder_keep_prob = 1.0
            decoder_keep_prob = 1.0
        self.encoder_keep_prob = encoder_keep_prob
        self.decoder_keep_prob = decoder_keep_prob

    """
    def update_decay_learning_rate(self, cur_step, cur_lr, lr_decay_scheme, num_train_steps):

        if lr_decay_scheme == "luong10":
            start_decay_step = int(num_train_steps / 2)
            remain_steps = num_train_steps - start_decay_step
            decay_steps = int(remain_steps / 10)  # decay 10 times
            decay_factor = 0.5
        elif lr_decay_scheme == "luong234":
            start_decay_step = int(num_train_steps * 2 / 3)
            remain_steps = num_train_steps - start_decay_step
            decay_steps = int(remain_steps / 4)  # decay 4 times
            decay_factor = 0.5
        else: # do not decay
            start_decay_step = num_train_steps
            decay_steps = 0
            decay_factor = 1.0

        print(" [*] lr_decay_scheme=%s, start_decay_step=%d, decay_steps %d, decay_factor %g" %
              (lr_decay_scheme, start_decay_step, decay_steps, decay_factor))

        if cur_step < start_decay_step:
            ret = tf.convert_to_tensor(cur_lr, tf.float32)
        else:
            ret = tf.train.exponential_decay(
                cur_lr, (cur_step - start_decay_step), decay_steps, decay_factor, staircase=True)
        return ret
    """

    def print_model_size(self, tvars):
        total_parameters = 0
        for variable in tvars:
            # shape is an array of tf.Dimension
            shape = variable.get_shape()
            variable_parameters = 1
            for dim in shape:
                variable_parameters *= dim.value
            print("Trainable %s with %d parameters" % (variable.name, variable_parameters))
            total_parameters += variable_parameters
        print("Total number of trainable parameters is %d" % total_parameters)
        self.params_size = total_parameters

    @staticmethod
    def print_loss(prefix, loss_names, losses, postfix):
        template = "%s "
        for name in loss_names:
            template += "%s: " % name
            template += " %f | "
        template += "%s"
        template = re.sub(' +', ' ', template)
        avg_losses = []
        values = [prefix]

        for loss in losses:
            values.append(np.mean(loss))
            avg_losses.append(np.mean(loss))
        values.append(postfix)

        print(template % tuple(values))

    @abstractmethod
    def build_graph(self):
        pass

    @abstractmethod
    def prepare_io(self):
        pass

    @abstractmethod
    def prepare_loss(self):
        pass


    def prepare_optimizer(self, tvars, loss, optimizer_type, init_lr, grad_clip=5.0, grad_noise=0.0):

        if optimizer_type.lower() == "adam":
            print("Use AdamOptimizer")
            optimizer = tf.train.AdamOptimizer(init_lr)
        elif optimizer_type.lower() == "rmsprop":
            print("Use RMSPropOptimizer")
            optimizer = tf.train.RMSPropOptimizer(init_lr)
        elif optimizer_type.lower() == "sgd":
            print("Use SGDOptimizer")
            optimizer = tf.train.GradientDescentOptimizer(self.learning_rate)
        else:
            raise ValueError("optimizer_type only accept values from [Adam, rmsprop, sgd]")

        gradients = tf.gradients(
            loss, tvars, colocate_gradients_with_ops=True)

        clipped_grads, grad_norm = tf.clip_by_global_norm(gradients, tf.constant(grad_clip))

        # add gradient noise
        if grad_noise > 0:
            grad_std = tf.sqrt(grad_noise / tf.pow(1.0 + tf.to_float(self.global_step), 0.55))
            clipped_grads = [g + tf.truncated_normal(tf.shape(g), mean=0.0, stddev=grad_std) for g in clipped_grads]

        self.train_ops = optimizer.apply_gradients(zip(clipped_grads, tvars),
                                                   global_step=self.global_step)
        self.print_model_size(tvars)


    @abstractmethod
    def batch_2_feed(self, batch):
        pass


    def train(self, generator, sess, step_limits):
        local_t = 0
        start_time = time.time()

        output_names = ["loss"] + list(self.other_losses.keys())
        output_values = []
        for _ in range(len(output_names)):
            output_values.append(list())

        while True:
            try:
                batch = next(generator)
            except StopIteration:
                break
            if step_limits is not None and local_t >= step_limits:
                break
            feed_dict = self.batch_2_feed(batch)
            fetches = [self.train_ops, self.loss] + list(self.other_losses.values())
            results = sess.run(fetches, feed_dict)

            print(results)
            for idx, res in enumerate(results[1:]):
                output_values[idx].append(res)

            local_t += 1
            if local_t % 500 == 0:
                outs = [np.mean(item) for item in output_values]
                self.print_loss("", output_names, outs, "")

        epoch_time = time.time() - start_time
        print("\n")
        final_outs = [np.mean(item) for item in output_values]
        self.print_loss("Train Epoch:", output_names, final_outs, postfix="epoch_time: %.4f"%epoch_time)

        return output_names, final_outs


    def test(self, generator, sess):
        start_time = time.time()
        output_names = ["loss"] + list(self.other_losses.keys())
        output_values = []
        for _ in range(len(output_names)):
            output_values.append(list())

        while True:
            try:
                batch = next(generator)
            except StopIteration:
                break
            feed_dict = self.batch_2_feed(batch)
            fetches = [self.loss] + list(self.other_losses.values())
            results = sess.run(fetches, feed_dict)

            for idx, res in enumerate(results):
                output_values[idx].append(res)

            #print(output_values)

        epoch_time = time.time() - start_time
        print("\n")
        final_outs = [np.mean(item) for item in output_values]
        self.print_loss("Test Results:", output_names, final_outs, postfix="test_time: %.4f"%epoch_time)
        return output_names, final_outs


import tensorflow as tf
from deepdialog.model.model_wrapper import TrainModel, EvalModel, InferModel
from deepdialog.model.kgcvae.kgcvae import KGCVAEModel


def create_vhred_train_model(dataset, dataloader, train_config, scope):
    train_data = dataloader.get_train_data()
    train_label_list = dataloader.get_train_label_list()
    train_generator = dataloader.get_generator(train_data, train_label_list)

    word2id = dataset.get_word2id()
    id2word = dataset.get_id2word()
    act_size = dataset.get_act_size()
    topic_size = dataset.get_label_size_by_name("topic")

    graph = tf.Graph()
    with graph.as_default():
        model = KGCVAEModel(train_config,
                            word2id,
                            id2word,
                            topic_size,
                            act_size,
                            train_generator,
                            mode=tf.contrib.learn.ModeKeys.TRAIN,
                            scope=scope,
                            dtype=tf.float32)
    return TrainModel(
        graph=graph,
        model=model,
        iterator=train_generator)


def create_vhred_test_model(dataset, dataloader, test_config, scope):
    test_data = dataloader.get_test_data()
    test_label_list = dataloader.get_test_label_list()
    test_generator = dataloader.get_generator(test_data, test_label_list)

    word2id = dataset.get_word2id()
    id2word = dataset.get_id2word()
    act_size = dataset.get_act_size()
    topic_size = dataset.get_label_size_by_name("topic")

    graph = tf.Graph()
    with graph.as_default():
        test_model = KGCVAEModel(
            test_config,
            word2id,
            id2word,
            topic_size,
            act_size,
            test_generator,
            mode=tf.contrib.learn.ModeKeys.EVAL,
            scope=scope,
            dtype=tf.float32)

    return EvalModel(
        graph=graph,
        model=test_model,
        iterator=test_generator)


def create_vhred_infer_model():
    pass



import tensorflow as tf
from deepdialog.model.base_model import BaseModel
from deepdialog.component.encoder import encoder, context_encoder
from deepdialog.component.decoder import decoder
from deepdialog.component import variational_network
from deepdialog.evaluation.evaluator import gaussian_kld
from deepdialog.loss_function.loss import sparse_softmax_cross_entropy_with_logits


class VHREDModel(BaseModel):
    def __init__(self, params, word2id, id2word, iterator, mode, scope=None, dtype=tf.float32):
        super().__init__(params, word2id, id2word, mode, scope)

        self.sos_id = word2id["<sos>"]
        self.eos_id = word2id["<eos>"]
        self.mode = mode
        # self.iterator = iterator

        with tf.name_scope("io"):
            self.encoder_inputs = tf.placeholder(dtype=tf.int32, shape=(None, None, None), name="encoder_inputs")
            self.encoder_lengths = tf.placeholder(dtype=tf.int32, shape=(None, None), name="encoder_lengths")
            self.decoder_inputs = tf.placeholder(dtype=tf.int32, shape=(None, None), name="decoder_inputs")
            self.decoder_lengths = tf.placeholder(dtype=tf.int32, shape=(None,), name="decoder_lengths")
            self.decoder_targets = tf.placeholder(dtype=tf.int32, shape=(None, None), name="decoder_targets")

            batch_size = tf.shape(self.encoder_inputs)[0]
            encoder_utt_num = tf.shape(self.encoder_inputs)[1]
            encoder_utt_len = tf.shape(self.encoder_inputs)[2]

        with tf.variable_scope(scope or "vhred", dtype=dtype):
            # response_encoder
            with tf.variable_scope("response_encoder"):
                if mode == tf.contrib.learn.ModeKeys.INFER:
                    response_states = None
                else:
                    _, response_states = encoder(self.mask_embeddings, self.decoder_targets,
                                                          self.decoder_lengths, params, self.encoder_keep_prob)
                # response_states: (batch_size, utt_enc_dim) * num_layers

            # utterance_encoder
            with tf.variable_scope("utterance_encoder"):
                utt_enc_inputs = tf.reshape(self.encoder_inputs,
                                            [batch_size * encoder_utt_num,
                                             encoder_utt_len])  # (batch_size*utt_num, utt_len)
                utt_enc_lens = tf.reshape(self.encoder_lengths, [batch_size * encoder_utt_num, ])
                utt_encoder_outputs, utt_encoder_states = encoder(self.mask_embeddings, utt_enc_inputs,
                                                                  utt_enc_lens, self.params,
                                                                  self.encoder_keep_prob)
                # utt_enc_states: (batch_size*utt_num, utt_enc_dim)* num_layers

            # context_encoder
            with tf.variable_scope("context_encoder"):
                ctx_enc_inputs = tf.layers.dense(inputs=utt_encoder_states,
                                                 units=params.ctx_enc_dim, activation=tf.tanh, name="utt_states2ctx_inputs")
                ctx_enc_inputs = tf.reshape(ctx_enc_inputs, [batch_size, encoder_utt_num, params.ctx_enc_dim])
                ctx_enc_lens = tf.fill([batch_size], encoder_utt_num)
                ctx_enc_outputs, ctx_enc_states = context_encoder(ctx_enc_inputs,
                                                                  ctx_enc_lens, params, self.encoder_keep_prob)
            # ctx_enc_states: (batch_size, ctx_enc_dim)

            # variation_network
            if mode == tf.contrib.learn.ModeKeys.TRAIN:
                prior_mu, prior_logvar = variational_network.prior(ctx_enc_states, params)
                posterior_mu, posterior_logvar = variational_network.posterior(ctx_enc_states, response_states, params)
                latent_sample = variational_network.sample_gaussian(posterior_mu, posterior_logvar)

            elif mode == tf.contrib.learn.ModeKeys.EVAL:
                prior_mu, prior_logvar = variational_network.prior(ctx_enc_states, params)
                posterior_mu, posterior_logvar = variational_network.posterior(ctx_enc_states, response_states, params)
                latent_sample = variational_network.sample_gaussian(prior_mu, prior_logvar)
            elif mode == tf.contrib.learn.ModeKeys.INFER:
                prior_mu, prior_logvar = variational_network.prior(ctx_enc_states, params)
                posterior_mu, posterior_logvar = None, None
                latent_sample = variational_network.sample_gaussian(prior_mu, prior_logvar)
            else:
                raise ValueError("valid mode")

            self.latent_sample = latent_sample
            concat_states = tf.concat([ctx_enc_states, latent_sample], -1)

            # decoder
            ctx_enc_outputs = tf.layers.dense(inputs=ctx_enc_outputs,
                                              units=params.hidden_units, activation=tf.tanh, name="ctx_outputs2decoder")
            concat_states = tf.layers.dense(inputs=concat_states,
                                            units=params.hidden_units, activation=tf.tanh, name="concat_states2decoder")
            logits, sample_id, final_decoder_states = decoder(
                self.mask_embeddings, ctx_enc_outputs, concat_states,
                self.encoder_lengths, self.decoder_inputs,
                self.decoder_lengths, self.mode, self.params, self.decoder_keep_prob)

            if mode == tf.contrib.learn.ModeKeys.TRAIN:
                self.predict_count = tf.reduce_sum(self.decoder_lengths)
                self.rc_loss = self._compute_loss(logits, self.decoder_targets, self.decoder_lengths)
                kl_weights = tf.minimum(tf.to_float(self.global_t) / params.full_kl_step, 1.0)
                self.kl_loss = tf.reduce_mean(gaussian_kld(posterior_mu, posterior_logvar, prior_mu, prior_logvar))
                self.elbo = self.rc_loss + kl_weights * self.kl_loss
                self.train_loss = self.elbo
                self.prepare_optimize(self.elbo)
            elif mode == tf.contrib.learn.ModeKeys.EVAL:
                self.predict_count = tf.reduce_sum(self.decoder_lengths)
                self.rc_loss = self._compute_loss(logits, self.decoder_targets, self.decoder_lengths)
                kl_weights = kl_weights = tf.constant(1.0)
                self.kl_loss = tf.reduce_mean(gaussian_kld(posterior_mu, posterior_logvar, prior_mu, prior_logvar))
                self.elbo = self.rc_loss + kl_weights * self.kl_loss
                self.eval_loss = self.elbo
            elif mode == tf.contrib.learn.ModeKeys.INFER:
                self.rc_loss = None
                self.kl_loss = None
                self.elbo = None
                self.infer_logits, self.sample_id = logits, sample_id
            else:
                raise ValueError("valid mode")

            self.saver = tf.train.Saver(tf.global_variables(), max_to_keep=3)

    @staticmethod
    def _compute_loss(logits, decoder_target, decoder_lengths):

        # print(logits)
        # print(decoder_target)

        cross_ent = tf.nn.sparse_softmax_cross_entropy_with_logits(
            labels=decoder_target, logits=logits)
        """
        max_time = dec_outputs.shape[1].value or tf.shape(dec_outputs)[1]
        batch_size = dec_outputs.shape[0].value or tf.shape(dec_outputs)[0]
        weights = tf.sequence_mask(dec_lengths, max_time, dtype=logits.dtype)
        """
        max_time = decoder_target.shape[1].value or tf.shape(decoder_target)[1]
        target_weights = tf.sequence_mask(decoder_lengths, max_time, dtype=logits.dtype)
        batch_size = tf.size(decoder_lengths)
        loss = tf.reduce_sum(cross_ent * target_weights) / tf.to_float(batch_size)
        return loss

    def train(self, sess, batch_data):
        assert self.mode == tf.contrib.learn.ModeKeys.TRAIN

        # print(len(batch_data))
        # print(batch_data)

        feed_dict = {
            self.encoder_inputs: batch_data[0],
            self.encoder_lengths: batch_data[1],
            self.decoder_inputs: batch_data[2],
            self.decoder_lengths: batch_data[3],
            self.decoder_targets: batch_data[4],
        }
        return sess.run([self.train_ops, self.train_loss,
                         self.elbo, self.rc_loss,
                         self.kl_loss, self.predict_count,
                         self.train_summary, self.global_t,
                         # self.batch_size,
                         # self.grad_norm,
                         self.learning_rate], feed_dict)

    def test(self, sess, batch_data):
        assert self.mode == tf.contrib.learn.ModeKeys.EVAL
        feed_dict = {
            self.encoder_inputs: batch_data[0],
            self.encoder_lengths: batch_data[1],
            self.decoder_inputs: batch_data[2],
            self.decoder_lengths: batch_data[3],
            self.decoder_targets: batch_data[4],
        }
        return sess.run([self.eval_loss, self.elbo, self.rc_loss, self.kl_loss,
                         self.predict_count], feed_dict)

    def interactive(self, sess, batch_data):
        assert self.mode == tf.contrib.learn.ModeKeys.INFER
        feed_dict = {
            self.encoder_inputs: batch_data[0],
            self.encoder_lengths: batch_data[1],
            self.decoder_inputs: batch_data[2],
            self.decoder_lengths: batch_data[3],
            self.decoder_targets: batch_data[4],
        }
        return sess.run([
            self.infer_logits, self.sample_id
        ], feed_dict)



from nltk.translate.bleu_score import sentence_bleu
import numpy as np
from nltk.util import ngrams
import math
# import tensorflow as tf


def average_length(source):
    """
    Calculate the average length among utterances.

    Parameters
    ----------
    source : list
        the element is a String that is an utterance

    Returns
    -------
    average_length : float
    """
    avg_len = float(sum([len(utt.strip().split()) for utt in source])) / len(source)
    return avg_len


def distinct_1(source):
    """
    Calculate the distinct-1 score.

    Distinct-1 means the quotient of the number of unique uni-gram divided by the number of
    all uni-gram. It illustrated the language diversity of the given text.

    Parameters
    ----------
    source : list
        the element is a String that is an utterance

    Returns
    -------
    distinct-1 : float
    """
    words = " ".join(source).split()
    num_distinct_words = len(set(words))
    return float(num_distinct_words) / len(words)


def distinct_2(source):
    """
    Calculate the distinct-2score.

    Distinct-1 means the quotient of the number of unique bi-gram divided by the number of
    all bi-gram. It illustrated the language diversity of the given text.

    Parameters
    ----------
    source : list
        the element is a String that is an utterance

    Returns
    -------
    distinct-2 : float
    """
    all_bigrams = []

    for line in source:
        line_list = line.split()
        bigrams = list(zip(line_list, line_list[1:]))
        all_bigrams.extend(list(bigrams))

    return len(set(all_bigrams)) / float(len(all_bigrams))


import os
import numpy as np
from itertools import chain
from deepdialog.data_handler.data_utils import clean_sentence_fn
from deepdialog.data_handler.data_utils import build_vocab_fn
from deepdialog.data_handler.data_utils import load_vocab_fn


__ALL__ = ['BaseDataset', 'DialogDataset']


class BaseDataset(object):
    """
    The :class:`BaseDataset` is the base class for loading and pre-processing text datasets.
    It supports reading text/text-related data from files of various formats, cleaning texts, and building vocabulary.

    The typical input for a :class:`BaseDataset` class is the path of dataset.
    Under the dataset_path, at least one file named ``texts.txt`` (or other formats) is supposed to be prepared.
    Within the ``texts.txt`` file, each line contains one data instance, which is a complete text
    (e.g., a dialogue, a document or a sentence).

    There are also cases where labels are required.
    :class:`BaseDataset` provides utilities for loading labels along with texts.
    Label files named ``xxx_labels.txt`` should be prepared in the same dataset_path directory.
    Within each ``xxx_labels.txt`` file, each line contains the label associating with the text data of the same line.
    Users could also provide more than one label files.

    When the given dataset_path contains no appropriate files, `FileNotFoundError` will be raised.

    Parameters
    ===========
    dataset_path: a string
        The directory path of dataset.
    max_vocab_size: an integer
        The maximum size of vocabulary to be built. If there are more than `max_vocab_size` words, words with low
        frequencies will be ignored. Default to be -1, which means no limits.
    label_name_list: a list
        The name list of text attributes to be considered.
        Each name ``xxx`` corresponds to one label file ``xxx_labels.txt``.
    prev_list: a list
        The prefix list of vocabulary. Usually they are manually-designated special tokens.
        Default to be an empty string.
    lowercase: boolean
        If true, transform all words into lowercase. Default to be False.
    rep_num: boolean
        If true, replace numbers with `<num>` token. Default to be False.
    rm_punc: boolean
        If true, remove all punctuation marks. Default to be False.
    """
    def __init__(self,
                 dataset_path,
                 max_vocab_size=-1,
                 prev_list=[],
                 label_name_list=[],
                 lowercase=True,
                 rep_num=True,
                 rm_punc=True):

        self._dataset_path = dataset_path
        self._max_vocab_size = max_vocab_size
        self._label_name_list = label_name_list
        self._lowercase = lowercase
        self._rep_num = rep_num
        self._rm_punc = rm_punc
        self._prev_list = prev_list

        self._data_size = -1
        self._vocab_size = -1

        # read texts
        self._raw_texts = self.read_texts()

        # clean texts
        self._cleaned_texts = self.clean_texts()

        # build vocabulary
        self._word2id, self._id2word = self.build_vocab()

        # read labels
        self._labels_list = self.read_labels()
        self._label2id_list, self._id2label_list = self.build_label_vocab()


    @property
    def dataset_path(self):
        """
        A string. Get current dataset path.
        """
        return self._dataset_path

    @property
    def max_vocab_size(self):
        """
        An integer. Get manually designated maximum size of vocabulary.
        """
        return self._max_vocab_size

    @property
    def label_name_list(self):
        """
        A list of strings. Get available attribute names.
        """
        return self._label_name_list

    @property
    def vocab_size(self):
        """
        An integer. Get the actual vocabulary size.
        """
        if self._vocab_size == -1:
            raise ValueError("No vocabulary built.")
        return self._vocab_size

    @property
    def data_size(self):
        """
        An integer. Get the number of total data instances.
        """
        return self._data_size

    def read_texts(self):
        """
        Read text data from `dataset_path/texts.txt`.

        Returns
        -------
        raw_texts: a list of strings.
            Each string is a line in texts.txt.
        """
        file_path = os.path.join(self._dataset_path, "texts.txt")
        with open(file_path, mode="r", encoding="utf-8") as file:
            raw_texts = file.read().split("\n")
        self._data_size = len(raw_texts)
        return raw_texts


    def clean_texts(self):
        """
        Pre-process raw texts, including transforming into lowercase, replacing numbers with `<num>`,
        replacing urls with `<url>`, and removing punctuations.

        Returns
        -------
        cleaned_texts: a list of strings
            The pre-processed texts.
        """
        cleaned_texts = [clean_sentence_fn(
            sample, lowercase=self._lowercase, rep_num=self._rep_num, rm_punc=self._rm_punc)
                         for sample in self._raw_texts]
        return cleaned_texts


    @property
    def raw_texts(self):
        """
        A list of strings. Get raw texts.
        """
        return self._raw_texts

    @property
    def cleaned_texts(self):
        """
        A list of strings. Get pre-processed texts.
        """
        return self._cleaned_texts

    def build_vocab(self, re_build=True):
        """
        Build the text vocabulary.
        If there's already a ``vocab.txt`` under the `dataset_path`, just load it.
        If not, build it, create ``vocab.txt`` and put it in `dataset_path`.

        Returns
        -------
        word2id: a dict
            Vocabulary that maps from indexes into words.
        id2word: a dict
            Vocabulary that maps from words into indexes.
        """

        vocab_file_path = os.path.join(self._dataset_path, "vocab.txt")
        if os.path.exists(vocab_file_path) and not re_build:
            word2id, id2word = load_vocab_fn(vocab_file_path)

        else:

            word_list = [sent.split() for sent in self._cleaned_texts]
            word_list = list(chain(*word_list))

            word2id, id2word = build_vocab_fn(word_list, self._max_vocab_size, self._prev_list)

            vocab_list = word2id.keys()
            with open(vocab_file_path, mode="w", encoding="utf-8") as file:
                for word in vocab_list:
                    file.write(word + "\n")

        self._vocab_size = len(word2id)
        return word2id, id2word


    @property
    def id2word(self):
        """
        Get a dictionary that maps from indexes into words.
        """
        return self._id2word

    @property
    def word2id(self):
        """
        Get a dictionary that maps from words into indexes.
        """
        return self._word2id

    def read_labels(self):
        """
        Read labels from {name}_labels.txt.

        Returns
        -------
        labels_list: a list of list of strings
            A list that contains multiple label list.
            Each label list is a list of strings, associating with one attribute.
            Each string represents a label value.
        """
        labels_list = []
        for name in self._label_name_list:
            label_file_path = os.path.join(self._dataset_path, name + "_labels.txt")
            with open(label_file_path, mode="r", encoding="utf-8") as file:
                labels = file.read().split("\n")
                assert len(labels) == self._data_size
                labels_list.append(labels)
        return labels_list


    def build_label_vocab(self, re_build=True):
        """
        Build the label vocabulary.
        If there's already a ``{label}_vocab.txt`` under the `dataset_path`, just load it.
        If not, build it, create ``{label}_vocab.txt`` and put it in `dataset_path`.

        Returns
        -------
        label2id_list: a list of dicts
            A list of label vocabularies.
            Each vocabulary associates with one attribute and maps from label values to indexes.
        id2label_list: a list of dicts
            A list of label vocabularies.
            Each vocabulary associates with one attribute and maps from indexes to label values.

        """
        label2id_list = []
        id2label_list = []
        for label_idx, name in enumerate(self._label_name_list):
            label_file_path = os.path.join(self._dataset_path, name + "_vocab.txt")
            if os.path.exists(label_file_path) and not re_build:
                label2id, id2label = load_vocab_fn(label_file_path)

            else:

                word_list = self._labels_list[label_idx]
                label2id, id2label = build_vocab_fn(word_list)

                label_list = label2id.keys()
                with open(label_file_path, mode="w", encoding="utf-8") as file:
                    for word in label_list:
                        file.write(word + "\n")

            label2id_list.append(label2id)
            id2label_list.append(id2label)

        return label2id_list, id2label_list


    @property
    def labels_list(self):
        return self._labels_list

    @property
    def label2id_list(self):
        return self._label2id_list

    @property
    def id2label_list(self):
        return self._id2label_list


class DialogDataset(BaseDataset):
    """
    The :class:`DialogDataset` is a class particularly for dialogue dataset processing.
    It inherits the parent class :class:`BaseDataset`.
    In addition to basic operation for text processing,
    :class:`DialogDataset` class also has dialogue-specific operations,
    including adding special tokens, splitting dialogues according to its word-utterance-dialogue hierarchy,
    and computing its relevant statistics.

    Parameters
    ===========
    dataset_path: a string
        The directory path of dataset.
    max_vocab_size: an integer
        The maximum size of vocabulary to be built. If there are more than `max_vocab_size` words, words with low
        frequencies will be ignored. Default to be -1, which means no limits.
    label_name_list: a list
        The name list of text attributes to be considered.
        Each name ``xxx`` corresponds to one label file ``xxx_labels.txt``.
    lowercase: boolean
        If true, transform all words into lowercase. Default to be False.
    rep_num: boolean
        If true, replace numbers with `<num>` token. Default to be False.
    rm_punc: boolean
        If true, remove all punctuation marks. Default to be False.

    """
    def __init__(self,
                 dataset_path,
                 turn_separator,
                 max_vocab_size=-1,
                 label_name_list=[],
                 lowercase=True,
                 rep_num=True,
                 rm_punc=True):
        BaseDataset.__init__(self, dataset_path, max_vocab_size, ["<pad>", "<unk>", "<sos>", "<eos>"],
                             label_name_list, lowercase, rep_num, rm_punc)

        self._turn_separator = turn_separator
        self._pad_id = self.word2id["<pad>"]
        self._unk_id = self.word2id["<unk>"]
        self._sos_id = self.word2id["<sos>"]
        self._eos_id = self.word2id["<eos>"]

        self._dialog_data = self.split_data()

        self.print_stats()

    @property
    def turn_separator(self):
        """
        A string. The special token that separates dialogue turns.
        It is supposed to present in the given dialogue dataset.
        """
        return self._turn_separator

    @property
    def pad_id(self):
        """
        An integer. The index number of `<pad>` token.
        """
        return self._pad_id

    @pad_id.setter
    def pad_id(self, value):
        if value != 0 or not isinstance(value, int):
            raise ValueError("pad_id is supposed to be integer 0.")
        self._pad_id = value

    @property
    def unk_id(self):
        """
        An integer. The index number of `<unk>` token.
        It represents all out-of-vocabulary words.
        """
        return self._unk_id

    @property
    def eos_id(self):
        """
        An integer. The index number of `<eos>` token.
        It represents the end-of-sentence.
        """
        return self._eos_id

    @property
    def sos_id(self):
        """
        An integer. The index number of `<sos>` token.
        It represents the start-of-sentence.
        """
        return self._sos_id

    def split_data(self):
        """
        Split dialogue data according to the word-utterance-dialog hierarchy.

        Returns
        -------
        dialog_data: A list of list of list of strings
            The dialogue data that have been split.
            The outer list represents the whole dataset.
            The medium list represents each dialogue, consisting of multiple utterances.
            The inner list represents each utterance, consisting of multiple words.
        """
        dialog_data = [[sent.split() for sent in dial.split(self._turn_separator)] for dial in self._cleaned_texts]

        return dialog_data

    @property
    def dialog_data(self):
        """
        Get the split dialog data.
        """
        return self._dialog_data

    def print_stats(self):
        """
        Compute and print basic statistics of the dialogue dataset,
        including total data size, average turns per dialogue, average lengths per turn, and vocabulary size.
        """
        print("===========================================")
        print("Statistics of dialogue dataset in %s." % self._dataset_path)
        print("===========================================")

        print("[*] Total data size: %d" % self._data_size)

        turns = [len(dial) for dial in self._dialog_data]
        print("[*] Max#turn: %d, Min#turn: %d, Avg turn per dialog: %.4f" % (np.max(turns), np.min(turns),
                                                                           np.mean(turns)))

        lens = [len(turn) for dial in self._dialog_data for turn in dial]
        print("[*] avg lengths per turn: %.4f" % np.mean(lens))

        print("[*] vocab size: %d" % self._vocab_size)


class TranslationDataset(BaseDataset):
    pass

class SummarizationDataset(BaseDataset):
    pass


from deepdialog.model import AbstractModel


class HDGMModel(AbstractModel):
    def build_graph(self):
        pass

    def prepare_io(self):
        pass

    def prepare_loss(self):
        pass

    def batch_2_feed(self, batch):
        pass

    def __init__(self):
        AbstractModel.__init__()


from abc import abstractmethod
from itertools import chain

import numpy as np
import sklearn


__ALL__ = ['BaseDataLoader', 'S2SDataLoader', 'M2MDataLoader', "M2SDataLoader"]


class BaseDataLoader(object):
    def __init__(self, dataset, split_ratio=None):

        if split_ratio is None:
            self._split_ratio = [8,1,1]
        else:
            self._split_ratio = split_ratio

        self._train_size = -1
        self._valid_size = -1
        self._test_size = -1

        self._train_dialog_data = None
        self._valid_dialog_data = None
        self._test_dialog_data = None

        self._train_label_list = []
        self._valid_label_list = []
        self._test_label_list = []


        self._train_tok_dialog_data = None
        self._valid_tok_dialog_data = None
        self._test_tok_dialog_data = None

        self._train_tok_label_list = []
        self._valid_tok_label_list = []
        self._test_tok_label_list = []


        self._dialog_data = dataset.dialog_data
        self._word2id = dataset.word2id
        self._unk_id = dataset.unk_id
        self._pad_id = dataset.pad_id
        self._sos_id = dataset.sos_id
        self._eos_id = dataset.eos_id
        self._label_data = dataset.labels_list
        self._label_name_list = dataset.label_name_list
        self._label2id_list = dataset.label2id_list
        self._data_size = dataset.data_size

        # if self._tokenized:

        self._tok_dialog_data = self.tokenize_dialogs(self._dialog_data)
        self._tok_label_data = self.tokenize_labels(self._label_data)

        self.split_train_valid_test()

    @property
    def train_size(self):
        return self._train_size

    @property
    def valid_size(self):
        return self._valid_size

    @property
    def test_size(self):
        return self._test_size

    @property
    def train_dialog_data(self):
        return self._train_dialog_data

    @property
    def valid_dialog_data(self):
        return self._valid_dialog_data

    @property
    def test_dialog_data(self):
        return self._test_dialog_data

    @property
    def train_label_list(self):
        return self._train_label_list

    @property
    def valid_label_list(self):
        return self._valid_label_list

    @property
    def test_label_list(self):
        return self._test_label_list

    @property
    def train_tok_dialog_data(self):
        return self._train_tok_dialog_data

    @property
    def valid_tok_dialog_data(self):
        return self._valid_tok_dialog_data

    @property
    def test_tok_dialog_data(self):
        return self._test_tok_dialog_data


    @property
    def train_tok_label_list(self):
        return self._train_tok_label_list

    @property
    def valid_tok_label_list(self):
        return self._valid_tok_label_list

    @property
    def test_tok_label_list(self):
        return self._test_tok_label_list

    """
    @property
    def tokenized(self):
        return self._tokenized
    """

    @property
    def split_ratio(self):
        return self._split_ratio

    def shuffle(self):
        outputs = sklearn.utils.shuffle(
            *([self._dialog_data, self._tok_dialog_data] + self._label_data + self._tok_label_data))
        self._dialog_data = outputs[0]
        self._tok_dialog_data = outputs[1]
        self._label_data = outputs[2:len(self._label_data) + 2]
        self._tok_label_data = outputs[len(self._label_data) + 2:]

        """
        if self._tokenized:
            outputs = sklearn.utils.shuffle(
                *([self._dialog_data, self._tok_dialog_data] + self._label_data + self._tok_label_data))
            self._dialog_data = outputs[0]
            self._tok_dialog_data = outputs[1]
            self._label_data = outputs[2:len(self._label_data)+2]
            self._tok_label_data = outputs[len(self._label_data)+2:]
        else:
            outputs = sklearn.utils.shuffle(*([self._dialog_data] + self._label_data))
            self._dialog_data = outputs[0]
            self._label_data = outputs[1:]
        """

    def tokenize_dialogs(self, dialog_data):
        ret = []
        for dial in dialog_data:
            tok_dial = [[self._word2id.get(word, self._unk_id) for word in utt] for utt in dial]
            ret.append(tok_dial)
        return ret

    def tokenize_labels(self, label_data):
        ret = []
        for idx, label2id in enumerate(self._label2id_list):
            cur_data = label_data[idx]
            tok_cur_data = [label2id.get(lab) for lab in cur_data]
            ret.append(tok_cur_data)
        return ret

    def split_train_valid_test(self):

        self.shuffle()

        self._train_size = int(self._data_size * self._split_ratio[0] / np.sum(self._split_ratio))
        self._valid_size = int(self._data_size * self._split_ratio[1] / np.sum(self._split_ratio))
        self._test_size = int(self._data_size * self._split_ratio[2] / np.sum(self._split_ratio))

        self._train_dialog_data = self._dialog_data[:self._train_size]
        self._valid_dialog_data = self._dialog_data[self._train_size:(self._train_size + self._valid_size)]
        self._test_dialog_data = self._dialog_data[(self._train_size + self._valid_size):]


        for labels in self._label_data:
            self._train_label_list.append(labels[:self._train_size])
            self._valid_label_list.append(labels[self._train_size:(self._train_size + self._valid_size)])
            self._test_label_list.append(labels[(self._train_size + self._valid_size):])

        """
        if self._tokenized:
            self._train_tok_dialog_data = self._tok_dialog_data[:self._train_size]
            self._valid_tok_dialog_data = self._tok_dialog_data[self._train_size:(self._train_size + self._valid_size)]
            self._test_tok_dialog_data = self._tok_dialog_data[(self._train_size + self._valid_size):]


            for labels in self._tok_label_data:
                self._train_tok_label_list.append(labels[:self._train_size])
                self._valid_tok_label_list.append(labels[self._train_size:(self._train_size + self._valid_size)])
                self._test_tok_label_list.append(labels[(self._train_size + self._valid_size):])
        """

        self._train_tok_dialog_data = self._tok_dialog_data[:self._train_size]
        self._valid_tok_dialog_data = self._tok_dialog_data[self._train_size:(self._train_size + self._valid_size)]
        self._test_tok_dialog_data = self._tok_dialog_data[(self._train_size + self._valid_size):]

        for labels in self._tok_label_data:
            self._train_tok_label_list.append(labels[:self._train_size])
            self._valid_tok_label_list.append(labels[self._train_size:(self._train_size + self._valid_size)])
            self._test_tok_label_list.append(labels[(self._train_size + self._valid_size):])


    def create_batch_generator(self, dialog_data, batch_size):

        all_lens = [len(dial) for dial in dialog_data]
        indexes = list(np.argsort(all_lens))
        batch_num = int(len(dialog_data) / batch_size)
        print("batch_num: %d" % batch_num)

        batch_indexes = []
        for idx in range(batch_num):
            start_index = idx * batch_size
            end_index = (idx + 1) * batch_size
            batch_indexes.append(indexes[start_index:end_index])

        for b_ids in batch_indexes:
            batch_data = [dialog_data[idx] for idx in b_ids]
            batch_input = self.process_batch_data(batch_data)
            yield batch_input


    @abstractmethod
    def process_batch_data(self, batch_data):
        pass

    def create_train_generator(self, batch_size):
        return self.create_batch_generator(self._train_tok_dialog_data, batch_size)

    def create_test_generator(self, batch_size):
        return self.create_batch_generator(self._test_tok_dialog_data, batch_size)

    def create_valid_generator(self, batch_size):
        return self.create_batch_generator(self._valid_tok_dialog_data, batch_size)


class S2SDataLoader(BaseDataLoader):

    def __init__(self, dataset, max_utt_len, max_dial_size, split_ratio):
        BaseDataLoader.__init__(self, dataset, split_ratio)

        self._max_utt_len = max_utt_len
        self._max_dial_size = max_dial_size

    def process_batch_data(self, batch_data):
        encoder_inputs = []
        encoder_lengths = []
        decoder_inputs = []
        decoder_lengths = []
        decoder_targets = []

        enc_length = (self._max_dial_size - 1) * self._max_utt_len
        dec_length = self._max_utt_len

        for dial in batch_data:
            assert len(dial) >= self._max_dial_size
            dial = dial[:self._max_dial_size]

            context = dial[:-1]
            context = list(chain(*context))[:enc_length]
            context_len = len(context)
            enc_input = context + [self._pad_id] * (enc_length - len(context))

            response = dial[-1][:dec_length]
            dec_len = len(response)
            dec_input = [self._sos_id] + response[:-1] + [self._pad_id] * (dec_length - len(response))
            dec_target = response[1:] + [self._eos_id] + [self._pad_id] * (dec_length - len(response))

            encoder_inputs.append(enc_input)
            encoder_lengths.append(context_len)
            decoder_inputs.append(dec_input)
            decoder_lengths.append(dec_len)
            decoder_targets.append(dec_target)

        return np.array(encoder_inputs), np.array(encoder_lengths), np.array(decoder_inputs), \
               np.array(decoder_lengths), np.array(decoder_targets)



class M2SDataLoader(BaseDataLoader):
    def __init__(self, dataset, max_utt_len, max_dial_size, split_ratio):
        BaseDataLoader.__init__(self, dataset, split_ratio)

        self._max_utt_len = max_utt_len
        self._max_dial_size = max_dial_size

    def process_batch_data(self, batch_data):
        encoder_inputs = []
        encoder_lengths = []
        decoder_inputs = []
        decoder_lengths = []
        decoder_targets = []

        dec_length = self._max_utt_len

        for dial in batch_data:
            assert len(dial) >= self._max_dial_size
            dial = dial[:self._max_dial_size]

            context = dial[:-1]
            enc_utt_input = []
            enc_utt_len = []
            for utt in context:
                utt = utt[:min(len(utt), self._max_utt_len)]
                enc_utt_len.append(len(utt))
                padded_utt = utt + [self._pad_id] * (self._max_utt_len - len(utt))
                enc_utt_input.append(padded_utt)

            response = dial[-1][:dec_length]
            dec_len = len(response)
            dec_input = [self._sos_id] + response[:-1] + [self._pad_id] * (dec_length - len(response))
            dec_target = response[1:dec_length] + [self._eos_id] + [self._pad_id] * (dec_length - len(response))

            encoder_inputs.append(enc_utt_input)
            encoder_lengths.append(enc_utt_len)
            decoder_inputs.append(dec_input)
            decoder_lengths.append(dec_len)
            decoder_targets.append(dec_target)

        return np.array(encoder_inputs), np.array(encoder_lengths), np.array(decoder_inputs), \
               np.array(decoder_lengths), np.array(decoder_targets)



class M2MDataLoader(BaseDataLoader):
    def __init__(self, dataset, max_utt_len, max_dial_size, split_ratio):
        BaseDataLoader.__init__(self, dataset, split_ratio)

        self._max_utt_len = max_utt_len
        self._max_dial_size = max_dial_size

    def process_batch_data(self, batch_data):
        pass














"""


class KGCVAEDataLoader(BaseDataLoader):

    def __init__(self, config, cleaned_texts, label_list, word2id, id2word, label2id_list, id2label_list,
                 label_name_list, acts, act2id, id2act):
        self._acts = acts
        self._act2id = act2id
        self._id2act = id2act
        self._acts = self._tokenize_acts(self._acts)
        super().__init__(config, cleaned_texts, label_list, word2id, id2word, label2id_list, id2label_list,
                         label_name_list)
        self._split_acts()

    def _shuffle_data(self):
        if len(self._label_list) != 0:
            output = sklearn.utils.shuffle(*([self._dialog_data] + [self._acts] + self._label_list))
            self._dialog_data = output[0]
            self._acts = output[1]
            self._label_list = output[2:]
        else:
            output = sklearn.utils.shuffle(*([self._dialog_data] + [self._acts] + self._label_list))
            self._dialog_data = output[0]
            self._acts = output[1]

    def get_generator(self, dialog_data, label_list):
        if len(label_list) != 0:
            assert len(dialog_data) == len(label_list[0])
        def generate():
            all_lens = [len(dial) for dial in dialog_data]  # [2,3,4,2]
            indexes = list(np.argsort(all_lens))  # [2,1,0,3]
            batch_num = int(len(dialog_data) / self._batch_size)

            batch_indexes = []
            for idx in range(batch_num):
                start_index = idx * self._batch_size
                end_index = (idx + 1) * self._batch_size
                batch_indexes.append(indexes[start_index:end_index])  # [2,1][0,3]

            for b_ids in batch_indexes:
                batch_data = [dialog_data[idx] for idx in b_ids]
                batch_acts = [self._acts[idx] for idx in b_ids]
                batch_labels = []
                for labels in label_list:
                    batch_labels.append([labels[idx] for idx in b_ids])

                batch_input = self._create_batch(batch_data, batch_labels, batch_acts)
                yield batch_input
        return generate

    def _create_batch(self, batch_data, batch_labels, batch_acts):
        encoder_inputs = []
        encoder_lengths = []
        decoder_inputs = []
        decoder_lengths = []
        decoder_targets = []
        acts = []
        floors = []

        dec_length = self._max_utterance_length
        assert "topic" in self._label_name_list
        label_idx = self._label_name_list.index("topic")
        topics = batch_labels[label_idx]

        for i in range(len(batch_data)):

            dial = batch_data[i]
            assert len(dial) >= self._max_dialog_size
            dial = dial[:self._max_dialog_size]

            context = dial[:-1]
            enc_input = []
            context_len = []
            floor = [0, 1] * len(context)
            floor = floor[:len(context)]
            for utt in context:
                utt = utt[:min(len(utt), self._max_utterance_length)]
                context_len.append(len(utt))
                padded_utt = utt + [self._pad_id] * (self._max_utterance_length - len(utt))
                enc_input.append(padded_utt)

            response = dial[-1][:(dec_length - 1)]
            dec_len = len(response) + 1
            dec_input = [self._sos_id] + response + [self._pad_id] * (dec_length - 1 - len(response))
            dec_target = response[:(dec_length - 1)] + [self._eos_id] + [self._pad_id] * (
                    dec_length - 1 - len(response))

            encoder_inputs.append(enc_input)
            encoder_lengths.append(context_len)
            decoder_inputs.append(dec_input)
            decoder_lengths.append(dec_len)
            decoder_targets.append(dec_target)
            acts.append(batch_acts[i][self._max_dialog_size-1])
            floors.append(floor)

        return [np.array(encoder_inputs), np.array(encoder_lengths), np.array(decoder_inputs),
                np.array(decoder_lengths), np.array(decoder_targets), np.array(topics),
                np.array(acts), np.array(floors)]

    def _tokenize_acts(self, acts):
        tokenized_acts = [[self._act2id.get(act)
                           for act in dialog] for dialog in acts]
        return tokenized_acts

    def _split_acts(self):

        self._train_acts = self._acts[:self._train_valid_split]
        self._valid_acts = self._acts[self._train_valid_split:self._valid_test_split]
        self._test_acts = self._acts[self._valid_test_split:]

"""


import tensorflow as tf
from deepdialog.model import AbstractModel
from deepdialog.component.encoder import Encoder
from deepdialog.component.decoder import Decoder

from deepdialog.utils.register import Registers


@Registers.model.register("Seq2SeqModel")
class Seq2SeqModel(AbstractModel):

    def __init__(self, params, mode, scope):
        AbstractModel.__init__(self, params, mode, scope)

        # define losses
        self.logits = None
        self.predict_word_count = None

    def prepare_io(self):
        self.encoder_inputs = tf.placeholder(dtype=tf.int32, shape=(None, None), name="encoder_inputs")
        self.encoder_lengths = tf.placeholder(dtype=tf.int32, shape=(None,), name="encoder_lengths")
        self.decoder_inputs = tf.placeholder(dtype=tf.int32, shape=(None, None), name="decoder_inputs")
        self.decoder_lengths = tf.placeholder(dtype=tf.int32, shape=(None,), name="decoder_lengths")
        self.decoder_targets = tf.placeholder(dtype=tf.int32, shape=(None, None), name="decoder_targets")

    def build_graph(self):
        with tf.variable_scope(self.scope or "seq2seq", dtype=tf.float32):
            self.seq2seq_encoder = Encoder(
                self.embeddings, self.params, self.encoder_keep_prob)

            encoder_outputs, encoder_states = self.seq2seq_encoder(self.encoder_inputs, self.encoder_lengths)

            self.seq2seq_decoder = Decoder(
                self.embeddings, self.mode, self.params, self.decoder_keep_prob)

            self.logits, sample_id, final_decoder_states = self.seq2seq_decoder(
                encoder_outputs, encoder_states, self.encoder_lengths,
                self.decoder_inputs, self.decoder_lengths)

    def prepare_loss(self):
        if self.mode == tf.contrib.learn.ModeKeys.INFER:
            self.loss = None
            self.other_losses = {}
        else: # train/eval

            """
            self.max_decoder_length = tf.reduce_max(self.decoder_lengths, name="max_decoder_length")
            self.mask = tf.sequence_mask(self.decoder_lengths, self.max_decoder_length, dtype=tf.float32, name='masks')
            loss = tf.contrib.seq2seq.sequence_loss(
                logits=self.logits, targets=self.decoder_targets, weights=self.mask)
            """
            cross_ent = tf.nn.sparse_softmax_cross_entropy_with_logits(
                labels=self.decoder_targets, logits=self.logits)

            max_time = self.decoder_targets.shape[1].value or tf.shape(self.decoder_targets)[1]
            target_weights = tf.sequence_mask(self.decoder_lengths, max_time, dtype=self.logits.dtype)
            batch_size = tf.size(self.decoder_lengths)
            loss = tf.reduce_sum(cross_ent * target_weights) / tf.to_float(batch_size)
            self.loss = loss
            self.other_losses["word_count"] = tf.reduce_sum(self.decoder_lengths)
            self.other_losses["nll"] = tf.reduce_sum(cross_ent * target_weights)
            self.other_losses["ppl"] = \
                tf.exp(tf.reduce_sum(cross_ent * target_weights)/tf.to_float(tf.reduce_sum(self.decoder_lengths)))

    def batch_2_feed(self, batch):
        encoder_inputs, encoder_lengths, decoder_inputs, decoder_lengths, decoder_targets = batch

        feed_dict = {
            self.encoder_inputs: encoder_inputs,
            self.encoder_lengths: encoder_lengths,
            self.decoder_inputs: decoder_inputs,
            self.decoder_lengths: decoder_lengths,
            self.decoder_targets: decoder_targets
        }
        return feed_dict


import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
import pandas as pd
from sklearn import manifold

def boxplot(value_matrix, labels, save_file):
    """
    show and save a boxplot to compare a measurement among models

    Parameters
    ----------
    value_matrix : list
        the element is a list that is the measurement values of a model
    labels : list
        the element is a String that is the name of model
    save_file : String
        the name of the file to be saved

    Returns
    -------
    boxplot : plt.Figure
    """
    values = np.transpose(np.array(value_matrix))
    values = pd.DataFrame(values, columns=labels)
    values.boxplot()
    fig = plt.gcf()
    plt.savefig(save_file)
    plt.show()
    plt.close()
    return fig


def kdeplot(value_matrix, labels, save_file):
    """
    show and save a boxplot to compare a measurement among models

    Parameters
    ----------
    value_matrix : list
        the element is a list that is the measurement values of a model
    labels : list
        the element is a String that is the name of model
    save_file : String
        the name of the file to be saved

    Returns
    -------
    kdeplot : plt.Figure
    """
    for i in range(len(value_matrix)):
        value = np.array(value_matrix[i])
        sns.kdeplot(value, bw=.2, label=labels[i])
    fig = plt.gcf()
    plt.savefig(save_file)
    plt.show()
    plt.close()
    return fig


def vector_cluster_scatter(vector_list, label_list, save_file):
    """
    show and save a scatter that shows the cluster of vectors

    Parameters
    ----------
    vector_list : list
        the element is a list that is a vector
    label_list : list
        the element is a String that is a label of a vector
    save_file : String
        the name of the file to be saved

    Returns
    -------
    scatter : plt.Figure
    """
    vector_list = np.array(vector_list)
    # standardization
    mu = np.mean(vector_list, axis=0)
    sigma = np.std(vector_list, axis=0)
    vector_list = (vector_list - mu) / sigma

    # tsne
    tsne = manifold.TSNE(n_components=2, init='pca', perplexity=5, random_state=0)
    tsne_vectors = tsne.fit_transform(vector_list)

    # DBSCAN
    # todo 调整参数
    # dbscan_vectors = DBSCAN(eps=5, min_samples=10).fit(tsne_vectors)
    # dbscan_labels = dbscan_vectors.labels_
    # print(dbscan_labels)
    # print(sum([1 for i in dbscan_labels if i == -1]))

    data_dict = dict()
    for i in range(len(label_list)):
        # if dbscan_labels[i] == -1:
        #     continue
        vector = tsne_vectors[i]
        label = label_list[i]
        if label in data_dict:
            data_dict[label].append(vector)
        else:
            data_dict[label] = [vector]
    for label, vectors in data_dict.items():
        vectors = np.array(vectors)
        plt.scatter(vectors[:, 0], vectors[:, 1], label=label)
    fig = plt.gcf()
    plt.savefig(save_file)
    plt.show()
    plt.close()
    return fig


def percent_stack_bar(vector_list, label_list, models_name, save_file):
    """
    show and save percent stack bar to compare some measurements among models

    Parameters
    ----------
    vector_list : list
        the element is a list that is scores of some measurement of a model
    label_list : list
        the element is a String that is the name of a measurement, len(label_list) == vector_list.shape[1]
    models_name : list
        the element is a String that is the name of a model, len(label_list) == vector_list.shape[0]
    save_file : String
        the name of the file to be saved

    Returns
    -------
    bar : plt.Figure
    """
    vectors = np.array(vector_list)
    m, n = vectors.shape
    percent = vectors / np.reshape(np.sum(vectors, axis=-1), [m, 1])
    accumulate = np.zeros(m)
    for i in range(n):
        plt.barh(y=models_name, width=percent[:, i], left=accumulate, label=label_list[i])
        accumulate += percent[:, i]
    ax = plt.gca()
    ax.set_xlim(0, 1)

    fig = plt.gcf()
    plt.savefig(save_file)
    plt.show()
    plt.close()
    return fig


def gantt_bar(vector_list, label_list, models_name, save_file):
    """
    show and save percent stack bar to compare some measurements among models

    Parameters
    ----------
    vector_list : list
        the element is a list that is scores of some measurement of a model
    label_list : list
        the element is a String that is the name of a measurement, len(label_list) == vector_list.shape[1]
    models_name : list
        the element is a String that is the name of a model, len(label_list) == vector_list.shape[0]
    save_file : String
        the name of the file to be saved

    Returns
    -------
    bar : plt.Figure

    """
    vectors = np.array(vector_list)
    m, n = vectors.shape
    percent = vectors / np.reshape(np.sum(vectors, axis=-1), [m, 1])
    middle_measurement_index = int(n / 2)
    middle_measurement_value = np.zeros(m)
    for i in range(middle_measurement_index):
        middle_measurement_value += percent[:, i]
    middle_measurement_value += percent[:, middle_measurement_index] / 2

    max_middle_value = max(middle_measurement_value)
    accumulate = max_middle_value - middle_measurement_value
    for i in range(n):
        plt.barh(y=models_name, width=percent[:, i], left=accumulate, label=label_list[i])
        accumulate += percent[:, i]
    fig = plt.gcf()
    plt.savefig(save_file)
    plt.show()
    plt.close()
    return fig


def knowledge_graph():
    # https: // zhuanlan.zhihu.com / p / 36700425
    pass


def self_attention_heatmap():
    pass


def dynamic_hidden_state():
    pass


import tensorflow as tf

from deepdialog.component.base import StochasticComponent
from deepdialog.distribution.distribution import MultiVarDiagonalGaussian

__ALL__ = ["PriorNet", "PosteriorNet"]


class PriorNet(StochasticComponent):
    """
    The :class:`PriorNet` class is the network for inferring prior distribution.
    It defines the prior distribution of a variable and supports stochastic sampling operation.
    """
    def __init__(self, latent_dim, n_samples=None, name="prior_net"):

        self._latent_dim = latent_dim
        self._n_samples = n_samples

        StochasticComponent.__init__(self, name, self._n_samples, dtype=tf.float32)

    @property
    def latent_dim(self):
        """The latent dimension of latent variable."""
        return self._latent_dim

    def __call__(self, context_states):
        with tf.variable_scope("prior_net"):
            prior_mulogvar = tf.layers.dense(inputs=context_states,
                                             units=self._latent_dim * 2,
                                             activation=tf.tanh, name="ctx_states2prior_network")
            prior_mu, prior_logvar = tf.split(prior_mulogvar, 2, axis=1)

            self._distribution = MultiVarDiagonalGaussian(
                prior_mu, prior_logvar, tf.float32, "prior_dist")
            z = self._distribution.sample(self._n_samples)
            return z


class PosteriorNet(StochasticComponent):
    """
    The :class:`PosteriorNet` class is the network for inferring posterior distribution.
    It defines the posterior distribution of a variable and supports stochastic sampling operation.
    """
    def __init__(self, latent_dim, n_samples=None, name="posterior_net"):

        self._latent_dim = latent_dim
        self._n_samples = n_samples

        StochasticComponent.__init__(self, name, self._n_samples, dtype=tf.float32)

    @property
    def latent_dim(self):
        """The latent dimension of latent variable."""
        return self._latent_dim

    def __call__(self, context_states, response_states):

        with tf.variable_scope("posterior_net"):
            posterior_input = tf.concat([context_states, response_states], 1)
            posterior_mulogvar = tf.layers.dense(inputs=posterior_input,
                                                 units=self._latent_dim * 2,
                                                 activation=tf.tanh, name="ctx_states2posterior_network")
            posterior_mu, posterior_logvar = tf.split(posterior_mulogvar, 2, axis=1)
            self._distribution = MultiVarDiagonalGaussian(
                posterior_mu, posterior_logvar, tf.float32, "posterior_dist")
            z = self._distribution.sample(self._n_samples)
            return z


import importlib


class Register(object):
    def __init__(self, registry_name):
        self._dict = {}
        self._name = registry_name

    def __setitem__(self, key, value):
        if not callable(value):
            raise Exception(f"Value of a Registry must be a callable!\nValue: {value}")
        if key is None:
            key = value.__name__
        if key in self._dict:
            print("Key %s already in registry %s." % (key, self._name))
        self._dict[key] = value

    def register(self, target):
        """Decorator to register a function or class."""

        def add(key, value):
            self[key] = value
            return value

        if callable(target):
            # @reg.register
            return add(None, target)
        # @reg.register('alias')
        return lambda x: add(target, x)

    def __getitem__(self, key):
        return self._dict[key]

    def __contains__(self, key):
        return key in self._dict

    def keys(self):
        """key"""
        return self._dict.keys()


class Registers(object):
    def __init__(self):
        raise RuntimeError("Registries is not intended to be instantiated")
    model = Register('model')


MODEL_MODULES = ["seq2seq", "hred"]
ALL_MODULES = [("deepdialog.model", MODEL_MODULES)]


def _handle_errors(errors):
    """Log out and possibly reraise errors during import."""
    if not errors:
        return
    for name, err in errors:
        print("Module {} import failed: {}".format(name, err))


def import_all_modules_for_register(custom_module_paths=None):
    """Import all modules for register."""
    modules = []
    print(ALL_MODULES)
    for base_dir, mod in ALL_MODULES:
        for name in mod:
            full_name = base_dir + "." + name
            modules.append(full_name)
    print(custom_module_paths)
    if isinstance(custom_module_paths, list):
        modules += custom_module_paths
    errors = []
    print(modules)
    for module in modules:
        try:
            importlib.import_module(module)
            print("successful load %s." % module)
        except ImportError as error:
            errors.append((module, error))
    _handle_errors(errors)



from nltk.translate.bleu_score import sentence_bleu
import numpy as np
from nltk.util import ngrams
import math
import tensorflow as tf


def Hw(source, token_prob, ngram):
    """
    Calculate the average trigram word entropy.

    This measurement is to evaluate the informativeness of the response (contrast with
    the general dull and ’safe’ responses).

    In particular, for a word :math:`w_n` in a response :math:`U` , the trigram word
    entropy is defined as:

    :math:`H(w_n) = -p(w_n|w_n-2, w_n-1)log(p(w_n|w_n-2, w_n-1))`

    References:
        https://dl.acm.org/doi/10.1145/3178876.3186077
        https://github.com/chenhongshen/HVMN/blob/master/model/Evaluation/entutil.py

    Parameters
    ----------
    source : list
        the element is a String that is an utterance and ends with "<eos>"
    token_prob : dict
        probabilities of unigram or bigram or trigram
    ngram : int
        1 for unigram, 2 for bigram, 3 fro trigram

    Returns
    -------
    average_word_entropy : float
    """
    word_entropy = []
    for line in source:
        if ngram == 1:
            word_list = line.split()
        elif ngram == 2:
            word_list = ["<sos>"] + line.split()
        elif ngram == 3:
            word_list = ["<sos>", "<sos>"] + line.split()
        else:
            assert ValueError("invalid ngram")

        if ngram == 1:
            ngram_list = word_list
        else:
            ngram_list = [" ".join(ngram_tuple) for ngram_tuple in ngrams(word_list, ngram)]
        for ngram_token in ngram_list:
            if ngram_token in token_prob:
                prob = token_prob[ngram_token]
            else:
                prob = token_prob["<new>"]
            word_entropy.append(-prob * math.log(prob, 2))
    return np.mean(word_entropy)


def perplexity(source_prob, target, target_mask, vocab_size):
    """
    Calculate the perplexity.

    In information theory, perplexity is a measurement of how well a probability
    distribution or probability model predicts a sample. It may be used to compare
    probability models. A low perplexity indicates the probability distribution is good
    at predicting the sample.

    The perplexity of a discrete probability distribution p is defined as:

    :math:`PPL = exp(-sum(p(x) * log(p(x))))`

    References:
        https://en.wikipedia.org/wiki/Perplexity

    Parameters
    ----------
    source_prob : tf.tensor
        probabilities of word generated from decoder, the shape is (batch_size * sequence_length * vocab_size)
    target : tf.tensor
        the tokenized response, the shape is (batch_size * sequence_length)
    target_mask : tf.tensor
        it implys the real length of target, the element is 0 or 1, 0 means no word, 1 means a word
    vocab_size : int
        the size of vocabulary

    Returns
    -------
    perplexity : tf.tensor
    """
    onehot_target = tf.one_hot(target, vocab_size)
    cross_entropy = -tf.reduce_sum(onehot_target * tf.log(source_prob), axis=-1)
    target_mask = tf.cast(target_mask, dtype=tf.float32)
    total_words = tf.reduce_sum(target_mask) + tf.constant(1e-10)
    avg_cross_entropy = tf.reduce_sum(cross_entropy * target_mask) / total_words
    return tf.exp(avg_cross_entropy)


def BERTScore(source, target):
    """
    Calculate the BERT score.

    BERTScore computes a similarity score for each token in the candidate sentence
    with each token in the reference sentence using contextual embeddings gotten from
    BERT model.

    References:
        https://arxiv.org/abs/1904.09675

    Parameters
    ----------
    source : list or String
    target : list or String

    Returns
    -------
    BERT_score : float
    """
    pass


from nltk.translate.bleu_score import sentence_bleu
import numpy as np
from nltk.util import ngrams
import math
import tensorflow as tf


def gaussian_kld(posterior_mu, posterior_logvar, prior_mu, prior_logvar):
    """
    Calculate the KL divergence between posterior distribution and prior distribution.

    In mathematical statistics, the KL(Kullback–Leibler) divergence is a measure of how
    one probability distribution is different from a second, reference probability
    distribution.

    In particular, if two distributions are gaussian distribution, the KL divergence can
    be calculated as:

    :math:`KL(p, q) = log(\delta_1) - log(\delta_2) + (\delta_1^2+(\mu_1 - \mu_2)^2)/(2*\delta_2^2) -0.5`

    where :math:`p` is the posterior distribution with :math:`\delta_1` as its standard
    deviation and :math:`\mu_1` as its mean, :math:`q` is the prior distribution with
    :math:`\delta_2` as its standard deviation and :math:`\mu_2` as its mean.

    References:
        https://arxiv.org/pdf/1606.05908.pdf
        https://www.cnblogs.com/yutingmoran/p/8631186.html

    Parameters
    ----------
    posterior_mu : tf.tensor
        the mean of posterior distribution, the shape is (batch_size * posterior_dim)
    posterior_logvar : tf.tensor
        the log of the covariance of posterior distribution, the shape is (batch_size * posterior_dim)
    prior_mu : tf.tensor
        the mean of prior distribution, the shape is (batch_size * prior_dim)
    prior_logvar : tf.tensor
        the log of the covariance of prior distribution, the shape is (batch_size * prior_dim)

    Returns
    -------
    KL_divergence : tf.tensor
        the shape is (batch_size,)
    """
    kld = -0.5 * tf.reduce_sum(1 + (posterior_logvar - prior_logvar)
                               - tf.div(tf.pow(prior_mu - posterior_mu, 2), tf.exp(prior_logvar))
                               - tf.div(tf.exp(posterior_logvar), tf.exp(prior_logvar)), reduction_indices=1)
    return kld


def standard_gaussian_kld(posterior_mu, posterior_logvar):
    """
    Calculate the KL divergence between posterior distribution and standard gaussian distribution.

    This is the special case of gaussian_kld() function where the prior distribution
    is standard gaussian distribution. Now the expression is:

    :math:`KL(p, q) = log(\delta_1) + (\delta_1^2+\mu_1^2)/2 -0.5`

    where :math:`p` is the posterior distribution with :math:`\delta_1` as its standard
    deviation and :math:`\mu_1` as its mean, :math:`q` is the standard gaussian
    distribution with 0 as mean, and 1 as standard deviation.

    References:
        https://arxiv.org/pdf/1606.05908.pdf
        https://www.cnblogs.com/yutingmoran/p/8631186.html

    Parameters
    ----------
    posterior_mu : tf.tensor
        the mean of posterior distribution, the shape is (batch_size * posterior_dim)
    posterior_logvar : tf.tensor
        the log of the covariance of posterior distribution, the shape is (batch_size * posterior_dim)

    Returns
    -------
    KL_divergence : tf.tensor
        the shape is (batch_size,)
    """
    kld = - 0.5 * tf.reduce_sum(1 + posterior_logvar
                                - tf.exp(posterior_logvar)
                                - tf.pow(posterior_mu, 2), reduction_indices=1)
    return kld


import time
import os
import tensorflow as tf
import numpy as np
from abc import abstractmethod

from deepdialog.utils.register import Registers


class Context(object):
    @classmethod
    def get_context_stack(cls):
        if not hasattr(cls, "_context_stack"):
            cls._context_stack = []
        return cls._context_stack

    @classmethod
    def get_context(cls):
        try:
            return cls.get_context_stack()[-1]
        except IndexError:
            raise RuntimeError("No contexts in the context_stack.")

    def __enter__(self):
        type(self).get_context_stack().append(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        type(self).get_context_stack().pop()




class AbstractFramework(object):

    def __init__(self, model_name, model_params, scope, retrain=True, ckp_dir=None, word2vec=None):

        print(type(self))
        self._model_name = model_name
        self._model_params = model_params
        self._scope = scope
        self._retrain = retrain

        self._sess = tf.Session()

        with tf.variable_scope(scope, reuse=None):
            self._train_model = Registers.model[model_name](params=model_params, mode="train", scope=scope)
        with tf.variable_scope(scope, reuse=True):
            self._eval_model = Registers.model[model_name](params=model_params, mode="eval", scope=scope)
        with tf.variable_scope(scope, reuse=True):
            self._infer_model = Registers.model[model_name](params=model_params, mode="tests", scope=scope)

        self._word2vec = word2vec
        self._ckp_dir = ckp_dir
        if not os.path.exists(self._ckp_dir):
            os.mkdir(self._ckp_dir)

        self.prepare_model()


    def prepare_model(self):
        if self._word2vec is not None:
            print("Initializing word2vec")
            self._sess.run(self._train_model.embeddings.assign(np.array(self._word2vec)))

        ckpt = tf.train.latest_checkpoint(self._ckp_dir)
        if ckpt and (not self._retrain):
            self._train_model.saver.restore(self._sess, ckpt)
            print("Restoring train model parameters from %s" % ckpt)
        else:
            print("Creating new train model with fresh parameters.")
            self._sess.run(tf.global_variables_initializer())
            self._sess.run(tf.tables_initializer())


    """
    def prepare_test_model(self):
        ckpt = tf.train.latest_checkpoint(self._ckp_dir)
        if ckpt:
            with self._eval_model.graph.as_default():
                self._eval_model.saver.restore(self._sess, ckpt)
                print("Restoring tests model parameters from %s" % ckpt)
        else:
            raise FileNotFoundError("Not Found ckpt files.")

    def prepare_infer_model(self):
        ckpt = tf.train.latest_checkpoint(self._ckp_dir)
        if ckpt:
            with self._infer_model.graph.as_default():
                self._infer_model.saver.restore(self._sess, ckpt)
                print("Restoring infer model parameters from %s" % ckpt)
        else:
            raise FileNotFoundError("Not Found ckpt files.")
    """

    @abstractmethod
    def train(self, dataloader, max_num_epoch, step_limits):
        pass

    @abstractmethod
    def test(self, dataloader):
        pass

    @abstractmethod
    def interactive(self):
        pass




class DialogFramework(AbstractFramework, Context):
    def __init__(self, model_name, model_params, scope, ckp_dir, retrain=False, word2vec=None):
        AbstractFramework.__init__(self, model_name, model_params, scope, retrain, ckp_dir, word2vec)

    def train(self, dataloader, max_num_epoch, step_limits=5000):
        start_time = time.time()
        best_valid_loss = np.inf
        checkpoint_path_name = os.path.join(self._ckp_dir, self._train_model.__class__.__name__ + ".ckpt")
        for epoch in range(max_num_epoch):

            cur_lr = self._sess.run(self._train_model.learning_rate)
            cur_step = self._sess.run(self._train_model.global_step)
            print(" [*] Epoch: %d | Step: %d | Learning Rate: %f" % (epoch, cur_step, cur_lr))

            dataloader.shuffle()

            train_generator = dataloader.create_train_generator(self._model_params.batch_size)
            train_names, train_outputs = self._train_model.train(
                train_generator, self._sess, step_limits=step_limits)
            train_loss = train_outputs[0]

            valid_generator = dataloader.create_valid_generator(self._model_params.batch_size)
            valid_names, valid_outputs = self._eval_model.test(valid_generator, self._sess)
            valid_loss = valid_outputs[0]

            if valid_loss < best_valid_loss:
                print("[*] Save current model into %s." % checkpoint_path_name)
                self._train_model.saver.save(self._sess, checkpoint_path_name, global_step=epoch)
                best_valid_loss = valid_loss

            print("[*] train_loss: %.4f | valid_loss: %.4f | elapsed_time: %.4f" %
                  (train_loss, valid_loss, (time.time() - start_time)))
            print("\n")

        print("\n")
        print("Finish Training.")
        print("Best valid loss %.4f" % best_valid_loss)


    def test(self, dataloader):
        start_time = time.time()
        dataloader.shuffle()
        test_generator = dataloader.create_test_generator(batch_size=self._model_params.batch_size)
        eval_names, eval_outputs = self._eval_model.test(test_generator, self._sess)


    def interactive(self):
        pass






# todo to implement multi-task dialogue framework.
class MDialogFramework(AbstractFramework, Context):
    def __init__(self, model_name, model_config, scope, ckp_dir=None, word2vec=None):
        AbstractFramework.__init__(model_name, model_config, scope, ckp_dir, word2vec)
        pass

    def train(self, dataloader, max_num_epoch, step_limits):
        pass

    def test(self, dataloader):
        pass

    def interactive(self):
        pass


import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
import pandas as pd
from sklearn import manifold


def align_heatmap(weight_matrix, source, target, save_file, color="Blues"):
    """
    Show and save an align heatmap for two sentence.

    Parameters
    ----------
    weight_matrix : list
        the weight matrix for the heatmap
    source : String
        the horizontal axis, word number of source must equal to weight_matrix.shape[1]
    target : String
        the vertical axis, word number of target must equal to weight_matrix.shape[0]
    save_file : String
        the name of the file to be saved
    color : String
        the color of the heatmap, with "Blues" as default

    Returns
    -------
    heatmap : plt.Figure
    """
    s = source.split()
    t = target.split()
    weight_matrix = np.array(weight_matrix)
    assert weight_matrix.shape[0] == len(t)
    assert weight_matrix.shape[1] == len(s)
    weights = pd.DataFrame(weight_matrix, columns=s, index=t)
    # sns.heatmap(weights, cmap=sns.dark_palette('white')) # white and black map
    # sns.heatmap(weights,cmap=sns.color_palette('Blues'), linewidth=0.5) # blue map
    sns.heatmap(weights, cmap=sns.color_palette(color), linewidth=0.5)

    ax = plt.gca()
    ax.xaxis.set_ticks_position('bottom')
    plt.xticks(rotation=90)
    fig = plt.gcf()
    plt.savefig(save_file)
    plt.show()
    plt.close()
    return fig


def highlight_heatmap(weight_matrix, source, save_file):
    """
    show and save a highlighted heatmap for some sentences

    Parameters
    ----------
    weight_matrix : list
        the weight list for the heatmap
    source : list
        the sentences that will be highlighted
    save_file : String
        the name of the file to be saved

    Returns
    -------

    """

    part1 = """
        <html lang="en">
        <head>
        <meta http-equiv="content-type" content="text/html; charset=utf-8">
        <style>
        body {
        font-family: Sans-Serif;
        }
        </style>
        </head>
        <body>
        <h3>
        Heatmaps
        </h3>
        </body>
        <script>
        """
    part2 = """
        var color = "255,0,0";
        var ngram_length = 3;
        var half_ngram = 1;
        for (var k=0; k < any_text.length; k++) {
        var tokens = any_text[k].split(" ");
        var intensity = new Array(tokens.length);
        var max_intensity = Number.MIN_SAFE_INTEGER;
        var min_intensity = Number.MAX_SAFE_INTEGER;
        for (var i = 0; i < intensity.length; i++) {
        intensity[i] = 0.0;
        for (var j = -half_ngram; j < ngram_length-half_ngram; j++) {
        if (i+j < intensity.length && i+j > -1) {
        intensity[i] += trigram_weights[k][i + j];
        }
        }
        if (i == 0 || i == intensity.length-1) {
        intensity[i] /= 2.0;
        } else {
        intensity[i] /= 3.0;
        }
        if (intensity[i] > max_intensity) {
        max_intensity = intensity[i];
        }
        if (intensity[i] < min_intensity) {
        min_intensity = intensity[i];
        }
        }
        var denominator = max_intensity - min_intensity;
        for (var i = 0; i < intensity.length; i++) {
        intensity[i] = (intensity[i] - min_intensity) / denominator;
        }
        if (k%2 == 0) {
        var heat_text = "<p><br><b>Example:</b><br>";
        } else {
        var heat_text = "<b>Example:</b><br>";
        }
        var space = "";
        for (var i = 0; i < tokens.length; i++) {
        heat_text += "<span style='background-color:rgba(" + color + "," + intensity[i] + ")'>" + space + tokens[i] + "</span>";
        if (space == "") {
        space = " ";
        }
        }
        //heat_text += "<p>";
        document.body.innerHTML += heat_text;
        }
        </script>
        </html>"""
    putQuote = lambda x: "\"%s\"" % x
    textsString = "var any_text = [%s];\n" % (",".join(map(putQuote, source)))
    weightsString = "var trigram_weights = [%s];\n" % (",".join(map(str, weight_matrix)))
    with open(save_file, "w", encoding="utf-8") as f:
        f.write(part1)
        f.write(textsString)
        f.write(weightsString)
        f.write(part2)


def attention_heatmap(weight_list, source, save_file, color="Blues"):
    """
    show and save a attention weights heatmap for a sentences

    Parameters
    ----------
    weight_list : list
        the attention weights for the heatmap
    source : String
        the horizontal axis, word number of source must equal to len(weight_list)
    save_file : String
        the name of the file to be saved
    color : String
        the color of the heatmap, with "Blues" as default

    Returns
    -------
    heatmap : plt.Figure
    """
    s = source.split()
    word_num = len(s)
    assert len(weight_list) == len(s)
    weight_matrix = np.array([weight_list])

    plt.figure(figsize=(word_num, 2))

    weights = pd.DataFrame(weight_matrix, columns=s)
    # sns.heatmap(weights, cmap=sns.dark_palette('white')) # white and black map
    # sns.heatmap(weights,cmap=sns.color_palette('Blues'), linewidth=0.5) # blue map
    sns.heatmap(weights, cmap=sns.color_palette(color), linewidth=0.5)
    ax = plt.gca()
    ax.xaxis.set_ticks_position('bottom')
    # plt.xticks(labels=s)
    plt.xticks(rotation=90)
    plt.yticks([])
    fig = plt.gcf()
    plt.savefig(save_file)
    plt.show()
    plt.close()
    return fig


def double_attention_heatmap(first_weight_list, second_weight_matrix, sources, save_file, color="Blues"):
    """
    show and save a attention weights heatmap for a sentences

    Parameters
    ----------
    first_weight_list : list
        the attention weights in sentence level
    second_weight_matrix : list
        the element is a list that is the attention weights in word level
    sources : list
        the element is a String that is a sentence
    save_file : String
        the name of the file to be saved
    color : String
        the color of the heatmap, with "Blues" as default

    Returns
    -------
    heatmap : plt.Figure
    """
    sentence_num = len(first_weight_list)
    max_sentence_len = max([len(i) for i in second_weight_matrix])
    first_w = np.transpose(np.array([first_weight_list]))

    first_w = pd.DataFrame(first_w, index=list(range(1, sentence_num+1)))
    grid = plt.GridSpec(sentence_num, 1+max_sentence_len)
    plt.subplot(grid[:, 0])
    sns.heatmap(first_w, cmap=sns.color_palette(color), linewidth=0.5)
    # plt.figure(figsize=(2, sentence_num))
    plt.xticks([])
    # plt.legend('off')

    for i in range(sentence_num):
        s = sources[i].split()

        second_w = np.array([second_weight_matrix[i]])
        second_w = pd.DataFrame(second_w, columns=s)
        plt.subplot(grid[i, 1:])
        sns.heatmap(second_w, cmap=sns.color_palette(color), linewidth=0.5)
        ax = plt.gca()
        ax.xaxis.set_ticks_position('bottom')
        # plt.xticks(labels=s)
        plt.xticks(rotation=90)
        plt.yticks([])

    fig = plt.gcf()
    plt.savefig(save_file)
    plt.show()
    plt.close()
    return fig



import tensorflow as tf

from deepdialog.component.base import DeterministicComponent

__ALL__ = ['Discriminator']


class Discriminator(DeterministicComponent):
    """
    The :class:`Discriminator` class defines the component for classifying sentences.
    It is implemented through fully-connected layers.
    """
    def __init__(self, n_classes, dropout_rate, input_dim, hidden_units, name="discriminator"):
        self._n_classes = n_classes
        self._dropout_rate = dropout_rate
        self._hidden_units = hidden_units
        self._input_dim = input_dim

        self._input_shape = tf.TensorShape([None, self._input_dim])
        self._output_shape = tf.TensorShape([None, self._n_classes])
        DeterministicComponent.__init__(self, name, self._input_shape, self._output_shape, tf.float32)

    @property
    def n_classes(self):
        """The number of classes to be classified."""
        return self._n_classes

    @property
    def dropout_rate(self):
        """The dropout rate. It means how much ratio of values to be dropped at each layer."""
        return self._dropout_rate

    @property
    def input_dim(self):
        """The input dimension."""
        return self._input_dim

    @property
    def hidden_units(self):
        """The dimension of middle layer."""
        return self._hidden_units

    def _check_input_shape(self, inputs):
        """Check whether the inputs are of the right shape."""
        err_msg = "The shape of inputs should match `input_shape`."
        try:
            inputs.get_shape()[-1].assert_is_compatible_with(self._input_shape[-1])
        except ValueError:
            raise ValueError(
                err_msg + "(inputs: {} | input_shape: {})".format(inputs.get_shape(), self._input_shape))
        return inputs


    def __call__(self, inputs):

        inputs = self._check_input_shape(inputs)
        if self._hidden_units is not None:
            dis_fc1 = tf.layers.dense(
                inputs=inputs,units=self._hidden_units, activation=tf.tanh, name="discriminator_fc1")
        else:
            dis_fc1 = inputs
        if self._dropout_rate < 1.0:
            dis_fc1 = tf.nn.dropout(dis_fc1, self._dropout_rate)
        logits = tf.layers.dense(inputs=dis_fc1,
                                 units=self._n_classes, activation=None, name="discriminator_project")
        return logits





import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="python-deepdialog",
    version="0.0.1",
    author="Yanan Zheng",
    author_email="zyanan93@gmail.com",
    description="DeepDialog is a lightweight library for deep generative dialogue models.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://zheng-yanan.github.io/deepdialog/",
    packages=setuptools.find_packages(exclude=['tests', 'tests.*',
                                    'examples', 'examples.*',
                                               'docs', 'docs.*',
                                               'data', 'data.*']),
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        # 'Programming Language :: Python :: 2',
        # 'Programming Language :: Python :: 2.7',
        # 'Programming Language :: Python :: 3',
        # 'Programming Language :: Python :: 3.4',
        # 'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=['numpy>=1.17.4',
                      'nltk>=3.4.5',
                      'seaborn>=0.10.0',
                      'matplotlib>=3.2.0',
                      'pandas>=1.0.1',
                      'sklearn'],

    extras_require={
        'dev': [
            'Sphinx>=1.7.1',
            'sphinx_rtd_theme',
            'sphinxcontrib-bibtex>=0.3.6',
            'pep8',
            'scipy',
            'coverage',
            'mock'
        ],
        'examples': [
            'scipy',
            'matplotlib',
            'scikit-image',
            'progressbar2'
        ],
    },
)


import tensorflow as tf

__ALL__ = ['create_rnn_cell']


def create_rnn_cell(cell_type, hidden_units, num_layers, input_keep_prob=1.0, output_keep_prob=1.0, use_residual=False):
    """
    Utilities for creating RNN cell.

    Parameters
    ----------
    cell_type: string
        The type of RNN cell. could be either `lstm` or `gru`.
    hidden_units: integer
        The number of hidden units.
    num_layers: integer
        The number of layers of RNN.
    """
    def _new_cell():
        if cell_type == "lstm":
            return tf.contrib.rnn.LSTMCell(hidden_units)
        elif cell_type == "gru":
            return tf.contrib.rnn.GRUCell(hidden_units)
        else:
            raise ValueError("cell_type must be either lstm or gru")

    def _new_wrapper_cell(residual_connection=False):
        cell = _new_cell()
        if input_keep_prob < 1.0 or output_keep_prob < 1.0:
            cell = tf.contrib.rnn.DropoutWrapper(cell,
                                                 input_keep_prob=input_keep_prob, output_keep_prob=output_keep_prob)
        if residual_connection:
            cell = tf.contrib.rnn.ResidualWrapper(cell)
        return cell

    if num_layers > 1:
        cells = []
        for i in range(num_layers):
            is_residual = True if use_residual and i > 1 else False
            cells.append(_new_wrapper_cell(is_residual))
        return tf.contrib.rnn.MultiRNNCell(cells)
    else:
        return _new_wrapper_cell(residual_connection=False)



import tensorflow as tf

__ALL__ = ['create_attention']


def create_attention(attention_type, hidden_units, memory, memory_lengths):
    """
    Utilities for creating attention.

    Parameters
    ----------
    attention_type: string
        The type of attention. Currently for types are provided: [`luong`, `scaled_luong`, `bahdanau`,
        `normed_bahdanau`].
    hidden_units: integer
        The number of hidden dimensions.
    memory: a Tensor
        The inputs to be memorized.
    memory_lengths: a Tensor
        The memory lengths.
    """
    if attention_type == "luong":
        attention_mechanism = tf.contrib.seq2seq.LuongAttention(hidden_units, memory,
                                                                memory_sequence_length=memory_lengths)
    elif attention_type == "scaled_luong":
        attention_mechanism = tf.contrib.seq2seq.LuongAttention(hidden_units, memory,
                                                                memory_sequence_length=memory_lengths, scale=True)
    elif attention_type == "bahdanau":
        attention_mechanism = tf.contrib.seq2seq.BahdanauAttention(hidden_units, memory,
                                                                   memory_sequence_length=memory_lengths)
    elif attention_type == "normed_bahdanau":
        attention_mechanism = tf.contrib.seq2seq.BahdanauAttention(hidden_units, memory,
                                                                   memory_sequence_length=memory_lengths, normalize=True)
    else:
        raise ValueError("Unknown attention_type: %s" % attention_type)

    return attention_mechanism
